#!/usr/bin/env python3
"""Support for Tuya Fan."""
from __future__ import annotations

import json
import logging
from typing import Any

from homeassistant.components.fan import (
    DIRECTION_FORWARD,
    DIRECTION_REVERSE,
    DOMAIN as DEVICE_DOMAIN,
    SUPPORT_DIRECTION,
    SUPPORT_OSCILLATE,
    SUPPORT_PRESET_MODE,
    SUPPORT_SET_SPEED,
    FanEntity,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)


# Fan
# https://developer.tuya.com/en/docs/iot/f?id=K9gf45vs7vkge
DPCODE_SWITCH = "switch"
DPCODE_FAN_SPEED = "fan_speed_percent"
DPCODE_MODE = "mode"
DPCODE_SWITCH_HORIZONTAL = "switch_horizontal"
DPCODE_FAN_DIRECTION = "fan_direction"

TUYA_SUPPORT_TYPE = {"fs"}  # Fan


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya fan dynamically through tuya discovery."""
    print("fan init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya fan."""
        print("fan add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: list):
    """Set up Tuya Fan."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue
        entities.append(TuyaHaFan(device, device_manager))
    return entities


class TuyaHaFan(TuyaHaDevice, FanEntity):
    """Tuya Fan Device."""

    def set_preset_mode(self, preset_mode: str) -> None:
        """Set the preset mode of the fan."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_MODE, "value": preset_mode}]
        )

    def set_direction(self, direction: str) -> None:
        """Set the direction of the fan."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_FAN_DIRECTION, "value": direction}]
        )

    def turn_off(self, **kwargs: Any) -> None:
        """Turn the device off."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_SWITCH, "value": False}]
        )

    def turn_on(
        self,
        speed: str = None,
        percentage: int = None,
        preset_mode: str = None,
        **kwargs: Any,
    ) -> None:
        """Turn on the fan."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_SWITCH, "value": True}]
        )

    def oscillate(self, oscillating: bool) -> None:
        """Oscillate the fan."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id,
            [{"code": DPCODE_SWITCH_HORIZONTAL, "value": oscillating}],
        )

    # property
    @property
    def is_on(self) -> bool:
        """Return true if fan is on."""
        return self.tuya_device.status.get(DPCODE_SWITCH, False)

    @property
    def current_direction(self) -> str:
        """Return the current direction of the fan."""
        return (
            DIRECTION_FORWARD
            if self.tuya_device.status.get(DPCODE_FAN_DIRECTION)
            else DIRECTION_REVERSE
        )

    @property
    def oscillating(self) -> bool:
        """Return true if the fan is oscillating."""
        return self.tuya_device.status.get(DPCODE_SWITCH_HORIZONTAL, False)

    @property
    def preset_modes(self) -> list:
        """Return the list of available preset_modes."""
        data = json.loads(self.tuya_device.function.get(DPCODE_MODE, {}).values).get(
            "range"
        )
        return data

    @property
    def preset_mode(self) -> str:
        """Return the current preset_mode."""
        return self.tuya_device.status.get(DPCODE_MODE)

    @property
    def percentage(self) -> int:
        """Return the current speed."""
        if not self.is_on:
            return 0
        return self.tuya_device.status.get(DPCODE_FAN_SPEED, 0)

    @property
    def supported_features(self):
        """Flag supported features."""
        supports = 0
        if DPCODE_MODE in self.tuya_device.status:
            supports = supports | SUPPORT_PRESET_MODE
        if DPCODE_FAN_SPEED in self.tuya_device.status:
            supports = supports | SUPPORT_SET_SPEED
        if DPCODE_SWITCH_HORIZONTAL in self.tuya_device.status:
            supports = supports | SUPPORT_OSCILLATE
        if DPCODE_FAN_DIRECTION in self.tuya_device.status:
            supports = supports | SUPPORT_DIRECTION
        return supports



import asyncio
import datetime
import logging
import os

import pytest
import requests_mock
from fhempy.lib.pkg_installer import check_and_install_dependencies
from tests.utils import mock_fhem


def load_fixture(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path) as fdp:
        return fdp.read()


@pytest.fixture(autouse=True)
def mock_ring_requests():
    with requests_mock.Mocker() as mock:
        mock.post(
            "https://oauth.ring.com/oauth/token", text=load_fixture("ring_oauth.json")
        )
        mock.post(
            "https://api.ring.com/clients_api/session",
            text=load_fixture("ring_session.json"),
        )
        mock.get(
            "https://api.ring.com/clients_api/ring_devices",
            text=load_fixture("ring_devices.json"),
        )
        mock.get(
            "https://api.ring.com/clients_api/chimes/999999/health",
            text=load_fixture("ring_chime_health_attrs.json"),
        )
        mock.get(
            "https://api.ring.com/clients_api/doorbots/987652/health",
            text=load_fixture("ring_doorboot_health_attrs.json"),
        )
        mock.get(
            "https://api.ring.com/clients_api/doorbots/987652/history",
            text=load_fixture("ring_doorbots.json"),
        )
        mock.get(
            "https://api.ring.com/clients_api/dings/active",
            text=load_fixture("ring_ding_active.json"),
        )
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/floodlight_light_off",
            text="ok",
        )
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/floodlight_light_on",
            text="ok",
        )
        mock.put("https://api.ring.com/clients_api/doorbots/987652/siren_on", text="ok")
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/siren_off", text="ok"
        )
        mock.get(
            "https://api.ring.com/clients_api/dings/987654321/recording",
            text="ok",
        )
        yield mock


@pytest.mark.asyncio
async def test_login(mocker):
    mock_fhem.mock_module(mocker)
    testhash = {"NAME": "testdevice"}
    await check_and_install_dependencies("ring")
    from fhempy.lib.ring.ring import ring

    fhempy_device = ring(logging.getLogger(__name__))

    await fhempy_device.Define(
        testhash,
        [
            "testdevice",
            "PythonModule",
            "ring",
            "test@test.com",
            "Front Door",
        ],
        {},
    )

    await fhempy_device.Set(
        testhash,
        [
            "testdevice",
            "password",
            "asdfasdfasdf",
        ],
        {},
    )

    count = 0
    while True:
        if (
            "state" in mock_fhem.readings["testdevice"]
            and mock_fhem.readings["testdevice"]["state"] == "connected"
        ):
            break
        if count > 10:
            break
        count += 1
        await asyncio.sleep(0.1)

    # wait for the ding
    await asyncio.sleep(0.2)
    assert mock_fhem.readings["testdevice"]["state"] == "ding"
    assert mock_fhem.readings["testdevice"]["address"] == "123 Main St"
    assert mock_fhem.readings["testdevice"]["family"] == "doorbots"
    assert mock_fhem.readings["testdevice"]["device_id"] == "aacdef123"
    assert mock_fhem.readings["testdevice"]["id"] == 987652
    assert mock_fhem.readings["testdevice"]["model"] == "Doorbell Pro"
    assert mock_fhem.readings["testdevice"]["firmware"] == "1.4.26"
    assert mock_fhem.readings["testdevice"]["latitude"] == 12.0
    assert mock_fhem.readings["testdevice"]["longitude"] == -70.12345
    assert mock_fhem.readings["testdevice"]["kind"] == "lpd_v1"
    assert mock_fhem.readings["testdevice"]["name"] == "Front Door"
    assert mock_fhem.readings["testdevice"]["timezone"] == "America/New_York"
    assert mock_fhem.readings["testdevice"]["wifi_name"] == "ring_mock_wifi"
    assert mock_fhem.readings["testdevice"]["wifi_signal_strength"] == -58
    assert mock_fhem.readings["testdevice"]["wifi_signal_category"] == "good"
    assert mock_fhem.readings["testdevice"]["battery_life"] == 100
    assert mock_fhem.readings["testdevice"]["existing_doorbell_type"] == "Mechanical"
    assert mock_fhem.readings["testdevice"]["existing_doorbell_type_enabled"] == True
    assert mock_fhem.readings["testdevice"]["existing_doorbell_type_duration"] == None
    assert mock_fhem.readings["testdevice"]["subscribed"] == True
    assert mock_fhem.readings["testdevice"]["subscribed_motion"] == True
    assert mock_fhem.readings["testdevice"]["has_subscription"] == True
    assert mock_fhem.readings["testdevice"]["volume"] == 1
    assert mock_fhem.readings["testdevice"]["connection_status"] == "online"
    assert mock_fhem.readings["testdevice"]["last_recording_id"] == 987654321
    assert (
        mock_fhem.readings["testdevice"]["last_recording_url"]
        == "https://api.ring.com/clients_api/dings/987654321/recording?api_version=9"
    )
    assert mock_fhem.readings["testdevice"]["history_1_id"] == 987654321
    assert mock_fhem.readings["testdevice"]["history_1_kind"] == "motion"
    assert mock_fhem.readings["testdevice"]["history_1_answered"] == False
    assert mock_fhem.readings["testdevice"][
        "history_1_created_at"
    ] == datetime.datetime(2017, 3, 5, 15, 3, 40, tzinfo=datetime.timezone.utc)

    assert mock_fhem.readings["testdevice"]["alert_id"] == 123456789

    await fhempy_device.Undefine(testhash)



#!/usr/bin/env python3
"""Support for Tuya Climate."""

import json
import logging
from typing import List

from homeassistant.components.climate import (
    DOMAIN as DEVICE_DOMAIN,
    HVAC_MODE_OFF,
    SUPPORT_FAN_MODE,
    SUPPORT_SWING_MODE,
    SUPPORT_TARGET_HUMIDITY,
    SUPPORT_TARGET_TEMPERATURE,
    ClimateEntity,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.const import TEMP_CELSIUS, TEMP_FAHRENHEIT
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)


# Air Conditioner
# https://developer.tuya.com/en/docs/iot/f?id=K9gf46qujdmwb
DPCODE_SWITCH = "switch"
DPCODE_TEMP_SET = "temp_set"
DPCODE_TEMP_SET_F = "temp_set_f"
DPCODE_MODE = "mode"
DPCODE_HUMIDITY_SET = "humidity_set"
DPCODE_TEMP_UNIT_CONVERT = "temp_unit_convert"
DPCODE_FAN_SPEED_ENUM = "fan_speed_enum"

# swing flap switch
DPCODE_SWITCH_HORIZONTAL = "switch_horizontal"
DPCODE_SWITCH_VERTICAL = "switch_vertical"

# status
DPCODE_TEMP_CURRENT = "temp_current"
DPCODE_TEMP_CURRENT_F = "temp_current_f"
DPCODE_HUMIDITY_CURRENT = "humidity_current"

SWING_OFF = "swing_off"
SWING_VERTICAL = "swing_vertical"
SWING_HORIZONTAL = "swing_horizontal"
SWING_BOTH = "swing_both"

TUYA_HVAC_TO_HA = {
    "hot": "heat",
    "cold": "cool",
    "wet": "dry",
    "wind": "fan_only",
    "auto": "auto",
}

TUYA_SUPPORT_TYPE = {"kt", "qn"}  # Air conditioner  # Heater


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya climate dynamically through tuya discovery."""
    print("climate init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya climate."""
        print("climate add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: List):
    """Set up Tuya Climate."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue
        entities.append(TuyaHaClimate(device, device_manager))
    return entities


class TuyaHaClimate(TuyaHaDevice, ClimateEntity):
    """Tuya Switch Device."""

    target_temp = 0.0
    target_h = 30

    def set_hvac_mode(self, hvac_mode):
        """Set new target hvac mode."""
        commands = []
        if hvac_mode == HVAC_MODE_OFF:
            commands.append({"code": DPCODE_SWITCH, "value": False})
        else:
            commands.append({"code": DPCODE_SWITCH, "value": True})

        for tuya_mode, ha_mode in TUYA_HVAC_TO_HA.items():
            if ha_mode == hvac_mode:
                commands.append({"code": DPCODE_MODE, "value": tuya_mode})

        self.tuya_device_manager.sendCommands(self.tuya_device.id, commands)

    def set_fan_mode(self, fan_mode):
        """Set new target fan mode."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_FAN_SPEED_ENUM, "value": fan_mode}]
        )

    def set_humidity(self, humidity):
        """Set new target humidity."""
        response = self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_HUMIDITY_SET, "value": int(humidity)}]
        )
        if response.get("success", False):
            self.target_h = humidity

    def set_swing_mode(self, swing_mode):
        """Set new target swing operation."""
        commands = []
        if swing_mode == SWING_BOTH:
            commands = [
                {"code": DPCODE_SWITCH_VERTICAL, "value": True},
                {"code": DPCODE_SWITCH_HORIZONTAL, "value": True},
            ]
        elif swing_mode == SWING_HORIZONTAL:
            commands = [
                {"code": DPCODE_SWITCH_VERTICAL, "value": False},
                {"code": DPCODE_SWITCH_HORIZONTAL, "value": True},
            ]
        elif swing_mode == SWING_VERTICAL:
            commands = [
                {"code": DPCODE_SWITCH_VERTICAL, "value": True},
                {"code": DPCODE_SWITCH_HORIZONTAL, "value": False},
            ]
        else:
            commands = [
                {"code": DPCODE_SWITCH_VERTICAL, "value": False},
                {"code": DPCODE_SWITCH_HORIZONTAL, "value": False},
            ]

        self.tuya_device_manager.sendCommands(self.tuya_device.id, commands)

    def set_temperature(self, **kwargs):
        """Set new target temperature."""
        print("climate temp->", kwargs)
        code = (
            DPCODE_TEMP_SET
            if self.tuya_device.status.get(DPCODE_TEMP_UNIT_CONVERT) == "c"
            else DPCODE_TEMP_SET_F
        )
        response = self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": code, "value": int(kwargs["temperature"])}]
        )
        if response.get("success", False):
            self.target_temp = kwargs["temperature"]

    # property

    @property
    def temperature_unit(self) -> str:
        """Return true if fan is on."""
        if self.tuya_device.status.get(DPCODE_TEMP_UNIT_CONVERT) == "c":
            return TEMP_CELSIUS
        else:
            return TEMP_FAHRENHEIT

    @property
    def current_temperature(self) -> float:
        """Return the current temperature."""
        if self.tuya_device.status.get(DPCODE_TEMP_UNIT_CONVERT) == "c":
            return self.tuya_device.status.get(DPCODE_TEMP_CURRENT, 0)
        else:
            return self.tuya_device.status.get(DPCODE_TEMP_CURRENT_F, 0)

    @property
    def current_humidity(self) -> int:
        """Return the current humidity."""
        return int(self.tuya_device.status.get(DPCODE_HUMIDITY_CURRENT, 0))

    @property
    def target_temperature(self) -> float:
        """Return the temperature currently set to be reached."""
        return self.target_temp

    @property
    def target_temperature_high(self) -> float:
        """Return the upper bound target temperature."""
        if self.tuya_device.status.get(DPCODE_TEMP_UNIT_CONVERT) == "c":
            temp_value = json.loads(
                self.tuya_device.function.get(DPCODE_TEMP_SET, {}).values
            )
            return temp_value.get("max", 0)
        else:
            temp_value = json.loads(
                self.tuya_device.function.get(DPCODE_TEMP_SET_F, {}).values
            )
            return temp_value.get("max", 0)

    @property
    def target_temperature_low(self) -> float:
        """Return the lower bound target temperature."""
        if self.tuya_device.status.get(DPCODE_TEMP_UNIT_CONVERT) == "c":
            temp_value = json.loads(
                self.tuya_device.function.get(DPCODE_TEMP_SET, {}).values
            )
            return temp_value.get("min", 0)
        else:
            temp_value = json.loads(
                self.tuya_device.function.get(DPCODE_TEMP_SET_F, {}).values
            )
            return temp_value.get("min", 0)

    @property
    def target_temperature_step(self) -> float:
        """Return target temperature setp."""
        return 1

    @property
    def target_humidity(self) -> int:
        """Return target humidity."""
        return int(self.target_h)

    @property
    def hvac_mode(self) -> str:
        """Return hvac mode."""
        if not self.tuya_device.status.get(DPCODE_SWITCH):
            return HVAC_MODE_OFF

        return TUYA_HVAC_TO_HA[self.tuya_device.status.get(DPCODE_MODE)]

    @property
    def hvac_modes(self) -> List:
        """Return hvac modes for select."""
        modes = json.loads(self.tuya_device.function.get(DPCODE_MODE, {}).values).get(
            "range"
        )

        print("hvac_modes->", modes)
        hvac_modes = [HVAC_MODE_OFF]
        for tuya_mode, ha_mode in TUYA_HVAC_TO_HA.items():
            if tuya_mode in modes:
                hvac_modes.append(ha_mode)

        return hvac_modes

    @property
    def fan_mode(self) -> str:
        """Return fan mode."""
        return self.tuya_device.status.get(DPCODE_FAN_SPEED_ENUM)

    @property
    def fan_modes(self) -> List[str]:
        """Return fan modes for select."""
        data = json.loads(
            self.tuya_device.function.get(DPCODE_FAN_SPEED_ENUM, {}).values
        ).get("range")
        return data

    @property
    def swing_mode(self) -> str:
        """Return swing mode."""
        mode = 0
        if (
            DPCODE_SWITCH_HORIZONTAL in self.tuya_device.status
            and self.tuya_device.status.get(DPCODE_SWITCH_HORIZONTAL)
        ):
            mode += 1
        if (
            DPCODE_SWITCH_VERTICAL in self.tuya_device.status
            and self.tuya_device.status.get(DPCODE_SWITCH_VERTICAL)
        ):
            mode += 2

        if mode == 3:
            return SWING_BOTH
        elif mode == 2:
            return SWING_VERTICAL
        elif mode == 1:
            return SWING_HORIZONTAL
        else:
            return SWING_OFF

    @property
    def swing_modes(self) -> List:
        """Return swing mode for select."""
        return [SWING_OFF, SWING_HORIZONTAL, SWING_VERTICAL, SWING_BOTH]

    @property
    def supported_features(self):
        """Flag supported features."""
        supports = 0
        if (
            DPCODE_TEMP_SET in self.tuya_device.status
            or DPCODE_TEMP_SET_F in self.tuya_device.status
        ):
            supports = supports | SUPPORT_TARGET_TEMPERATURE
        if DPCODE_FAN_SPEED_ENUM in self.tuya_device.status:
            supports = supports | SUPPORT_FAN_MODE
        if DPCODE_HUMIDITY_SET in self.tuya_device.status:
            supports = supports | SUPPORT_TARGET_HUMIDITY
        if (
            DPCODE_SWITCH_HORIZONTAL in self.tuya_device.status
            or DPCODE_SWITCH_VERTICAL in self.tuya_device.status
        ):
            supports = supports | SUPPORT_SWING_MODE
        return supports



import asyncio
import functools

import spotipy
from spotipy.oauth2 import CacheFileHandler
from aiohttp import web
from pyppeteer import launch

from .. import fhem, utils
from ..generic import FhemModule


class spotify_connect_player(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        # Spotipy PKCE authenticator instance
        self.spotipy_pkce = None
        self.spotipy_scope = (
            ""
            # Spotify scopes: https://developer.spotify.com/documentation/general/guides/scopes/
            # Listening history
            + "user-read-recently-played "
            + "user-top-read "
            + "user-read-playback-position "
            # Spotify connect
            + "user-read-playback-state "
            + "user-modify-playback-state "
            + "user-read-currently-playing "
            # Playback
            + "streaming "
            # Playlist
            + "playlist-read-private "
            + "playlist-read-collaborative "
            # Follow
            + "user-follow-read "
            # Library
            + "user-library-read "
            # User
            + "user-read-email "
            + "user-read-private"
        )
        self.loggedin = False
        attr_config = {"player_name": {"default": "FHEM Web Player"}}
        self.set_attr_config(attr_config)

        self.set_config = {"start": {}, "stop": {}}
        self.set_set_config(self.set_config)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "disconnected")
        await fhem.readingsEndUpdate(hash, 1)
        await self.set_auth_url()
        self.create_async_task(self.connect_spotipy())

    async def set_auth_url(self):
        # not sure if this is useful...
        x = "e92855a009"
        y = "e74eb69ba6609d3bfd7d"
        z = 90 + 6
        handler = CacheFileHandler(cache_path=f".{self.hash['NAME']}_spotify_token")
        self.spotipy_pkce = spotipy.oauth2.SpotifyPKCE(
            f"{x}{y}{str(z)}",
            "https://europe-west1-fhem-ga-connector.cloudfunctions.net/codelanding/start",
            scope=self.spotipy_scope,
            cache_handler=handler,
        )
        url = self.spotipy_pkce.get_authorize_url()
        url = (
            '<html><a href="'
            + url
            + '" target="_blank">Connect Spotify account (new window/tab)</a><br></html>'
        )
        await fhem.readingsSingleUpdate(self.hash, "login", url, 1)

    async def set_authcode(self, hash, params):
        code = params["code"]
        self.create_async_task(self.handle_authcode(code))

    async def handle_authcode(self, code):
        access_token = await utils.run_blocking(
            functools.partial(
                self.spotipy_pkce.get_access_token, code=code, check_cache=False
            )
        )
        await self.connect_spotipy()

    async def connect_spotipy(self):
        if self.spotipy_pkce.get_cached_token() is None:
            await fhem.readingsSingleUpdate(self.hash, "state", "login required", 1)
            return

        if self.spotipy is None:
            self.spotipy = spotipy.Spotify(auth_manager=self.spotipy_pkce)
            user = self.spotipy.current_user()
            if user is not None:
                self.loggedin = True
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "user",
                    user["display_name"] + " (" + user["email"] + ")",
                    1,
                )
                await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)
                await self.update_devices()
                self.create_async_task(self.update_playback_loop())
                self.create_async_task(self.update_devices_loop())

    async def set_start(self, hash, params):
        self.create_async_task(self.run_spotify())

    async def run_spotify(self):
        async def handle(request):
            spotify_connect_code = """
                <script src="https://sdk.scdn.co/spotify-player.js"></script>
                <script>
                    window.onSpotifyWebPlaybackSDKReady = () => {
                    const token = '###SPOTIFY_ACCESS_TOKEN###';
                    const player = new Spotify.Player({
                        name: '###SPOTIFY_WEB_PLAYER_NAME###',
                        getOAuthToken: cb => { cb(token); }
                    });

                    // Error handling
                    player.addListener('initialization_error', ({ message }) => { console.error(message); });
                    player.addListener('authentication_error', ({ message }) => { console.error(message); });
                    player.addListener('account_error', ({ message }) => { console.error(message); });
                    player.addListener('playback_error', ({ message }) => { console.error(message); });

                    // Playback status updates
                    player.addListener('player_state_changed', state => { console.log(state); });

                    // Ready
                    player.addListener('ready', ({ device_id }) => {
                        console.log('Ready with Device ID', device_id);
                    });

                    // Not Ready
                    player.addListener('not_ready', ({ device_id }) => {
                        console.log('Device ID has gone offline', device_id);
                    });

                    // Connect to the player!
                    player.connect();
                    };
                </script>
            """
            access_token = self.spotipy_pkce.get_cached_token()
            if access_token is not None:
                spotify_connect_code = spotify_connect_code.replace(
                    "###SPOTIFY_ACCESS_TOKEN###", access_token["access_token"]
                )
                spotify_connect_code = spotify_connect_code.replace(
                    "###SPOTIFY_WEB_PLAYER_NAME###", self._attr_player_name
                )
            return web.Response(text=spotify_connect_code)

        # start aiohttp server and run chrome headless
        self.app = web.Application()
        self.app.add_routes([web.get("/", handle)])

        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, "localhost", 8080)
        await self.site.start()

        self.browser = launch(
            {
                "ignoreDefaultArgs": ["--mute-audio"],
                "executablePath": "/usr/bin/chromium",
            }
        )
        page = await self.browser.newPage()
        await page.goto("http://localhost:8080/")

    async def set_stop(self, hash, params):
        self.create_async_task(self.stop_spotify())

    async def stop_spotify(self):
        # stop aiohttp server and stop chrome headless
        await self.runner.cleanup()
        await self.browser.close()



import asyncio
import concurrent.futures
import logging
import threading
import traceback

from fhempy.lib.generic import FhemModule
from zeroconf import ServiceBrowser, Zeroconf

from .. import fhem
from ..core.zeroconf import zeroconf


class discover_mdns(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.loop = asyncio.get_event_loop()
        self.zeroconf = None
        self.hash = None
        self.browser = None

    # zeroconf callback
    def update_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        self.logger.debug("Service %s updated, service info: %s" % (name, info))
        res = asyncio.run_coroutine_threadsafe(self.foundDevice(name, info), self.loop)
        res.result()

    # zeroconf callback
    def remove_service(self, zeroconf, type, name):
        self.logger.debug("Service %s removed" % (name))

    # zeroconf callback
    def add_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        self.logger.debug("Service %s added, service info: %s" % (name, info))
        res = asyncio.run_coroutine_threadsafe(self.foundDevice(name, info), self.loop)
        res.result()

    async def foundDevice(self, name, info):
        if info is None:
            return
        try:

            def get_value(key):
                """Retrieve value and decode to UTF-8."""
                value = info.properties.get(key.encode("utf-8"))

                if value is None or isinstance(value, str):
                    return value
                return value.decode("utf-8")

            if info.type == "_googlecast._tcp.local.":
                # check if device exists already, if not commanddefine
                if not (
                    await fhem.checkIfDeviceExists(
                        self.hash,
                        "PYTHONTYPE",
                        "googlecast",
                        "CASTNAME",
                        get_value("fn"),
                    )
                ):
                    self.logger.debug("create device: " + get_value("fn"))
                    await fhem.CommandDefine(
                        self.hash,
                        get_value("md").replace(" ", "_")
                        + "_"
                        + get_value("fn").replace(" ", "_")
                        + " PythonModule googlecast '"
                        + get_value("fn")
                        + "'",
                    )
                else:
                    self.logger.debug(
                        "device " + get_value("fn") + " exists already, do not create"
                    )
            elif info.type == "_soundtouch._tcp.local.":
                if not (
                    await fhem.checkIfDeviceExists(
                        self.hash, "TYPE", "BOSEST", "DEVICEID", "0"
                    )
                ):
                    self.logger.debug("create bosest")
                    await fhem.CommandDefine(self.hash, "bosesystem BOSEST")
                else:
                    self.logger.debug("device BOSEST exists already, do not create")
            elif (
                info.type == "_http._tcp.local."
                and info.name == "fhempy._http._tcp.local."
            ):
                if not (
                    await fhem.checkIfDeviceExists(
                        self.hash, "TYPE", "BindingsIo", "IP", get_value("ip")
                    )
                ):
                    ip = get_value("ip")
                    port = get_value("port")
                    ipstr = ip.replace(".", "_")
                    await fhem.CommandDefine(
                        self.hash, f"fhempy_peer_{ipstr} BindingsIo {ip}:{port} Python"
                    )
            elif info.type == "_spotify-connect":
                if not (
                    await fhem.checkIfDeviceExists(
                        self.hash, "PYTHONTYPE", "spotify", "PYTHONTYPE", "spotify"
                    )
                ):
                    self.logger.debug("create spotify")
                    await fhem.CommandDefine(
                        self.hash, "spotify_connect PythonModule spotify"
                    )
                else:
                    self.logger.debug("device spotify exists already, do not create")
            else:
                return

            # wait for the devices to initialize
            await asyncio.sleep(10)
        except Exception as err:
            self.logger.error(traceback.print_exc())

    async def runZeroconfScan(self):
        # await here to finish define before zeroconf object is created
        await asyncio.sleep(1)
        self.zeroconf = zeroconf.get_instance(self.logger).get_zeroconf()
        listener = self
        services = [
            "_googlecast._tcp.local.",
            "_soundtouch._tcp.local.",
            "_http._tcp.local.",
            "_spotify-connect",
        ]
        self.browser = ServiceBrowser(self.zeroconf, services, listener)

    # FHEM
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        await fhem.readingsSingleUpdate(self.hash, "state", "active", 1)
        self.create_async_task(self.runZeroconfScan())

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon rc_SEARCH")

    # FHEM
    async def Undefine(self, hash):
        if self.browser:
            self.browser.cancel()
        await super().Undefine(hash)



# THANKS to https://github.com/AlexxIT/XiaomiGateway3 for the HASS implementation
# which I used as a base for the FHEM implementation

import asyncio
import functools
import logging

from .. import fhem, utils
from ..generic import FhemModule
from .core.gateway3 import Gateway3

DOMAINS = [
    "binary_sensor",
    "climate",
    "cover",
    "light",
    "remote",
    "sensor",
    "switch",
    "alarm_control_panel",
]


class xiaomi_gateway3(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.gw = None
        self.devices = {}
        self.fhempy_devices = {}

    @property
    def gateway3(self):
        return self.gw.gateway3

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self.hash = hash

        if len(args) < 5:
            return "Usage: define devname PythonModule xiaomi_gateway3 <IP> <TOKEN>"

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon mqtt")
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "disconnected", 1)

        logging.getLogger("fhempy.lib.xiaomi_gateway3.core.gateway3").setLevel(
            logging.ERROR
        )

        hash["HOST"] = args[3]
        hash["TOKEN"] = args[4]

        self.host = args[3]
        self.token = args[4]

        self.create_async_task(self.connect_gw())

        return ""

    async def register_device(self, fhempy_device, handler):
        did = fhempy_device.did
        self.fhempy_devices[did] = fhempy_device
        # wait till gateway is ready
        while self.gw is None:
            await asyncio.sleep(3)
        self.gw.add_update(did, handler)
        if did == "lumi.0":
            self.gw.add_stats(did, handler)
        if did in self.devices and "init" in self.devices[did]:
            await fhempy_device.initialize(self.devices[did])

    async def connect_gw(self):
        config = {"devices": {}}
        self.gw = FhempyGateway(self.logger, self.hash, self.host, self.token, config)
        # prepare domains
        for domain in DOMAINS:
            self.gw.add_setup(domain, self.create_device)
        # start check task
        self.create_async_task(self.is_connected())
        # run gateway
        self.gw.start()

    async def is_connected(self):
        while True:
            if await self.gw.is_connected():
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "connected", 1
                )
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "disconnected", 1
                )
            await asyncio.sleep(60)

    async def create_device(self, gw, device, type):
        self.logger.debug(f"Check if device {device['did']} exists")
        did = device["did"]
        self.devices[did] = device
        if did == "lumi.0":
            did = "lumi." + device["mac"]
        if not await fhem.checkIfDeviceExists(
            self.hash, "PYTHONTYPE", "xiaomi_gateway3_device", "DID", did
        ):
            devname = (
                "".join(filter(str.isalnum, device["model"])) + "_" + device["mac"]
            )
            await fhem.CommandDefine(
                self.hash,
                devname
                + " PythonModule xiaomi_gateway3_device "
                + self.hash["NAME"]
                + " "
                + did,
            )
        else:
            if "init" in device and did in self.fhempy_devices:
                if did == "lumi." + device["mac"]:
                    device["version"] = self.gw.version()
                await self.fhempy_devices[did].initialize(device)


class FhempyGateway:
    def __init__(self, logger, hash, host, token, config):
        self.gw = Gateway3(host, token, config)
        self.loop = asyncio.get_event_loop()

    @property
    def gateway3(self):
        return self.gw

    def add_setup(self, domain, handler):
        def setup(gw, device, attr):
            asyncio.run_coroutine_threadsafe(
                handler(gw, device, attr), self.loop
            ).result()

        self.gw.add_setup(domain, setup)

    def add_update(self, did, handler):
        def update(data):
            asyncio.run_coroutine_threadsafe(handler(data), self.loop).result()

        self.gw.add_update(did, update)

    def add_stats(self, ieee, handler):
        def stats(data):
            asyncio.run_coroutine_threadsafe(handler(data), self.loop).result()

        self.gw.add_stats(ieee, stats)

    def version(self):
        return self.gw.ver

    def start(self):
        self.gw.start()

    async def is_connected(self):
        return await utils.run_blocking(functools.partial(self.gw._check_port, 23))



#!/usr/bin/env python3
"""Tuya Home Assistant Base Device Model."""
from __future__ import annotations

from tuya_iot import TuyaDevice, TuyaDeviceManager

from .const import DOMAIN


class TuyaHaDevice:
    """Tuya base device."""

    def __init__(self, device: TuyaDevice, device_manager: TuyaDeviceManager):
        """Init TuyaHaDevice."""
        super().__init__()

        self.tuya_device = device
        self.tuya_device_manager = device_manager
        self.entity_id = f"tuya_v2.ty{self.tuya_device.id}"

    @staticmethod
    def remap(old_value, old_min, old_max, new_min, new_max):
        """Remap old_value to new_value."""
        new_value = ((old_value - old_min) / (old_max - old_min)) * (
            new_max - new_min
        ) + new_min
        return new_value

    @property
    def should_poll(self) -> bool:
        """Hass should not poll."""
        return False

    @property
    def unique_id(self) -> str | None:
        """Return a unique ID."""
        return f"ty{self.tuya_device.id}"

    @property
    def name(self) -> str | None:
        """Return Tuya device name."""
        return self.tuya_device.name

    @property
    def device_info(self):
        """Return a device description for device registry."""
        _device_info = {
            "identifiers": {(DOMAIN, f"{self.tuya_device.id}")},
            "manufacturer": "tuya",
            "name": self.tuya_device.name,
            "model": self.tuya_device.product_name,
        }
        return _device_info

    # @property
    # def icon(self) -> Optional[str]:
    #     """Return Tuya device icon."""
    #     cdn_url = 'https://images.tuyacn.com/'
    #     # TODO customize cdn url
    #     return cdn_url + self.tuyaDevice.icon

    @property
    def available(self) -> bool:
        """Return if the device is available."""
        return self.tuya_device.online



#!/usr/bin/env python3
"""Support for the Tuya lights."""

import json
import logging
from typing import Any, Dict, List, Tuple

from tuya_iot import TuyaDevice, TuyaDeviceManager

from homeassistant.components.light import (
    ATTR_BRIGHTNESS,
    ATTR_COLOR_TEMP,
    ATTR_HS_COLOR,
    DOMAIN as DEVICE_DOMAIN,
    SUPPORT_BRIGHTNESS,
    SUPPORT_COLOR,
    SUPPORT_COLOR_TEMP,
    LightEntity,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)


# Light(dj)
# https://developer.tuya.com/en/docs/iot/f?id=K9i5ql3v98hn3
DPCODE_SWITCH = "switch_led"
DPCODE_WORK_MODE = "work_mode"
DPCODE_BRIGHT_VALUE = "bright_value"
DPCODE_TEMP_VALUE = "temp_value"
DPCODE_COLOUR_DATA = "colour_data"

MIREDS_MAX = 500
MIREDS_MIN = 153

HSV_HA_HUE_MIN = 0
HSV_HA_HUE_MAX = 360
HSV_HA_SATURATION_MIN = 0
HSV_HA_SATURATION_MAX = 100

WORK_MODE_WHITE = "white"
WORK_MODE_COLOUR = "colour"

TUYA_SUPPORT_TYPE = {
    "dj",  # light
    "dd",  # Light strip
    "fwl",  # Ambient light
    "dc",
    "jsq",  # Humidifier's light
}

DEFAULT_HSV = {
    "h": {"min": 1, "scale": 0, "unit": "", "max": 360, "step": 1},
    "s": {"min": 1, "scale": 0, "unit": "", "max": 255, "step": 1},
    "v": {"min": 1, "scale": 0, "unit": "", "max": 255, "step": 1},
}


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya light dynamically through tuya discovery."""
    print("light init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya light."""
        print("light add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: List):
    """Set up Tuya Light device."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue
        entities.append(TuyaHaLight(device, device_manager))
    return entities


class TuyaHaLight(TuyaHaDevice, LightEntity):
    """Tuya light device."""

    dp_code_bright = DPCODE_BRIGHT_VALUE
    dp_code_temp = DPCODE_TEMP_VALUE
    dp_code_colour = DPCODE_COLOUR_DATA

    def __init__(self, device: TuyaDevice, deviceManager: TuyaDeviceManager):
        """Init TuyaHaLight."""
        super().__init__(device, deviceManager)

        for key in device.function:
            if key.startswith(DPCODE_BRIGHT_VALUE):
                self.dp_code_bright = key
            elif key.startswith(DPCODE_TEMP_VALUE):
                self.dp_code_temp = key
            elif key.startswith(DPCODE_COLOUR_DATA):
                self.dp_code_colour = key

    @property
    def is_on(self) -> bool:
        """Return true if light is on."""
        return self.tuya_device.status.get(DPCODE_SWITCH, False)

    def turn_on(self, **kwargs: Any) -> None:
        """Turn on or control the light."""
        commands = []
        print("light kwargs->", kwargs)
        if (
            ATTR_BRIGHTNESS not in kwargs
            and ATTR_HS_COLOR not in kwargs
            and ATTR_COLOR_TEMP not in kwargs
        ):
            commands += [{"code": DPCODE_SWITCH, "value": True}]

        if ATTR_BRIGHTNESS in kwargs:
            if self._work_mode().startswith(WORK_MODE_COLOUR):
                colour_data = self._get_hsv()
                v_range = self._tuya_hsv_v_range()
                # hsv_v = colour_data.get('v', 0)
                colour_data["v"] = self.remap(
                    kwargs[ATTR_BRIGHTNESS], 0, 255, v_range[0], v_range[1]
                )
                commands += [
                    {"code": self.dp_code_colour, "value": json.dumps(colour_data)}
                ]
            else:
                new_range = self._tuya_brightness_range()
                tuya_brightness = int(
                    self.remap(
                        kwargs[ATTR_BRIGHTNESS], 0, 255, new_range[0], new_range[1]
                    )
                )
                commands += [{"code": self.dp_code_bright, "value": tuya_brightness}]
                # commands += [{'code': DPCODE_WORK_MODE, 'value': self._work_mode()}]

        if ATTR_HS_COLOR in kwargs:
            colour_data = self._get_hsv()
            # hsv h
            colour_data["h"] = kwargs[ATTR_HS_COLOR][0]
            # hsv s
            ha_s = kwargs[ATTR_HS_COLOR][1]
            s_range = self._tuya_hsv_s_range()
            colour_data["s"] = self.remap(
                ha_s,
                HSV_HA_SATURATION_MIN,
                HSV_HA_SATURATION_MAX,
                s_range[0],
                s_range[1],
            )
            # hsv v
            ha_v = self.brightness
            v_range = self._tuya_hsv_v_range()
            colour_data["v"] = self.remap(ha_v, 0, 255, v_range[0], v_range[1])

            commands += [
                {"code": self.dp_code_colour, "value": json.dumps(colour_data)}
            ]
            if self.tuya_device.status[DPCODE_WORK_MODE] != "colour":
                commands += [{"code": DPCODE_WORK_MODE, "value": "colour"}]

        if ATTR_COLOR_TEMP in kwargs:
            # temp color
            new_range = self._tuya_temp_range()
            color_temp = self.remap(
                self.max_mireds - kwargs[ATTR_COLOR_TEMP] + self.min_mireds,
                self.min_mireds,
                self.max_mireds,
                new_range[0],
                new_range[1],
            )
            commands += [{"code": self.dp_code_temp, "value": int(color_temp)}]

            # brightness
            ha_brightness = self.brightness
            new_range = self._tuya_brightness_range()
            tuya_brightness = self.remap(
                ha_brightness, 0, 255, new_range[0], new_range[1]
            )
            commands += [{"code": self.dp_code_bright, "value": int(tuya_brightness)}]

            if self.tuya_device.status[DPCODE_WORK_MODE] != "white":
                commands += [{"code": DPCODE_WORK_MODE, "value": "white"}]

        self.tuya_device_manager.sendCommands(self.tuya_device.id, commands)

    def turn_off(self, **kwargs: Any) -> None:
        """Instruct the light to turn off."""
        commands = [{"code": DPCODE_SWITCH, "value": False}]
        self.tuya_device_manager.sendCommands(self.tuya_device.id, commands)

    # LightEntity

    @property
    def brightness(self):
        """Return the brightness of the light."""
        old_range = self._tuya_brightness_range()
        brightness = self.tuya_device.status.get(self.dp_code_bright, 0)

        print(
            "brightness id->",
            self.tuya_device.id,
            "work_mode->",
            self._work_mode(),
            "; check true->",
            self._work_mode().startswith(WORK_MODE_COLOUR),
        )

        if self._work_mode().startswith(WORK_MODE_COLOUR):
            colour_data = json.loads(
                self.tuya_device.status.get(self.dp_code_colour, 0)
            )
            v_range = self._tuya_hsv_v_range()
            hsv_v = colour_data.get("v", 0)
            return int(self.remap(hsv_v, v_range[0], v_range[1], 0, 255))
        else:
            return int(self.remap(brightness, old_range[0], old_range[1], 0, 255))

    def _tuya_brightness_range(self) -> Tuple[int, int]:
        if self.dp_code_bright not in self.tuya_device.status:
            return 0, 255

        bright_value = json.loads(
            self.tuya_device.function.get(self.dp_code_bright, {}).values
        )
        return bright_value.get("min", 0), bright_value.get("max", 255)

    @property
    def hs_color(self):
        """Return the hs_color of the light."""
        colour_data = json.loads(self.tuya_device.status.get(self.dp_code_colour, 0))
        s_range = self._tuya_hsv_s_range()
        return colour_data.get("h", 0), self.remap(
            colour_data.get("s", 0),
            s_range[0],
            s_range[1],
            HSV_HA_SATURATION_MIN,
            HSV_HA_SATURATION_MAX,
        )

    @property
    def color_temp(self):
        """Return the color_temp of the light."""
        new_range = self._tuya_temp_range()
        tuya_color_temp = self.tuya_device.status.get(self.dp_code_temp, 0)
        ha_color_temp = (
            self.max_mireds
            - self.remap(
                tuya_color_temp,
                new_range[0],
                new_range[1],
                self.min_mireds,
                self.max_mireds,
            )
            + self.min_mireds
        )
        return ha_color_temp

    @property
    def min_mireds(self):
        """Return color temperature min mireds."""
        return MIREDS_MIN

    @property
    def max_mireds(self):
        """Return color temperature max mireds."""
        return MIREDS_MAX

    def _tuya_temp_range(self) -> Tuple[int, int]:
        temp_value = json.loads(
            self.tuya_device.function.get(self.dp_code_temp, {}).values
        )
        return temp_value.get("min", 0), temp_value.get("max", 255)

    def _tuya_hsv_s_range(self) -> Tuple[int, int]:
        colour_data = self._tuya_hsv_function()
        s = colour_data.get("s")
        return s.get("min", 0), s.get("max", 255)

    def _tuya_hsv_v_range(self) -> Tuple[int, int]:
        colour_data = self._tuya_hsv_function()
        v = colour_data.get("v")
        return v.get("min", 0), v.get("max", 255)

    def _tuya_hsv_function(self):
        hsv_data = json.loads(
            self.tuya_device.function.get(self.dp_code_colour, {}).values
        )
        if hsv_data == {}:
            return DEFAULT_HSV
        else:
            return hsv_data

    def _work_mode(self) -> str:

        return self.tuya_device.status.get(DPCODE_WORK_MODE, "")

    def _get_hsv(self) -> Dict[str, int]:
        return json.loads(self.tuya_device.status[self.dp_code_colour])

    @property
    def supported_features(self):
        """Flag supported features."""
        supports = 0
        if self.dp_code_bright in self.tuya_device.status:
            supports = supports | SUPPORT_BRIGHTNESS
        if (
            self.dp_code_colour in self.tuya_device.status
            and len(self.tuya_device.status[self.dp_code_colour]) > 0
        ):
            supports = supports | SUPPORT_COLOR
        if self.dp_code_temp in self.tuya_device.status:
            supports = supports | SUPPORT_COLOR_TEMP
        return supports



"""Two classes for read Unqlite and SQLite DB files frow raw bytes. Default
python sqlite3 library can't read DB from memory.
"""


class Unqlite:
    page_size = 0
    pos = 0

    def __init__(self, raw: bytes):
        self.raw = raw
        self.read_db_header()

    @property
    def size(self):
        return len(self.raw)

    def read(self, length: int):
        self.pos += length
        return self.raw[self.pos - length:self.pos]

    def read_int(self, length: int):
        return int.from_bytes(self.read(length), 'big')

    def read_db_header(self):
        assert self.read(7) == b'unqlite', "Wrong file signature"
        assert self.read(4) == b'\xDB\x7C\x27\x12', "Wrong DB magic"
        creation_time = self.read_int(4)
        sector_size = self.read_int(4)
        self.page_size = self.read_int(4)
        assert self.read(6) == b'\x00\x04hash', "Unsupported hash"

    # def read_header2(self):
    #     self.pos = self.page_size
    #     magic_numb = self.read(4)
    #     hash_func = self.read(4)
    #     free_pages = self.read_int(8)
    #     split_bucket = self.read_int(8)
    #     max_split_bucket = self.read_int(8)
    #     next_page = self.read_int(8)
    #     num_rect = self.read_int(4)
    #     for _ in range(num_rect):
    #         logic_page = self.read_int(8)
    #         real_page = self.read_int(8)

    def read_cell(self):
        key_hash = self.read(4)
        key_len = self.read_int(4)
        data_len = self.read_int(8)
        next_offset = self.read_int(2)
        overflow_page = self.read_int(8)
        if overflow_page:
            self.pos = overflow_page * 0x1000 + 8
            data_page = self.read_int(8)
            data_offset = self.read_int(2)
            name = self.read(key_len)
            self.pos = data_page * 0x1000 + data_offset
            value = self.read(data_len)
        else:
            name = self.read(key_len)
            value = self.read(data_len)
        return name, value, next_offset

    def read_all(self) -> dict:
        result = {}

        page_offset = 2 * self.page_size
        while page_offset < self.size:
            self.pos = page_offset
            next_offset = self.read_int(2)
            while next_offset:
                self.pos = page_offset + next_offset
                k, v, next_offset = self.read_cell()
                result[k.decode()] = v.decode()
            page_offset += self.page_size

        return result


class SQLite:
    page_size = 0
    pos = 0

    def __init__(self, raw: bytes):
        self.raw = raw
        self.read_db_header()
        self.tables = self.read_page(0)

    @property
    def size(self):
        return len(self.raw)

    def read(self, length: int):
        self.pos += length
        return self.raw[self.pos - length:self.pos]

    def read_int(self, length: int):
        return int.from_bytes(self.read(length), 'big')

    def read_varint(self):
        result = 0
        while True:
            i = self.read_int(1)
            result += i & 0x7f
            if i < 0x80:
                break
            result <<= 7

        return result

    def read_db_header(self):
        assert self.read(16) == b'SQLite format 3\0', "Wrong file signature"
        self.page_size = self.read_int(2)

    def read_page(self, page_num: int):
        self.pos = 100 if page_num == 0 else self.page_size * page_num

        # B-tree Page Header Format
        page_type = self.read(1)

        if page_type == b'\x0D':
            return self._read_leaf_table(page_num)
        elif page_type == b'\x05':
            return self._read_interior_table(page_num)
        else:
            raise NotImplemented

    def _read_leaf_table(self, page_num: int):
        first_block = self.read_int(2)
        cells_num = self.read_int(2)
        cells_pos = self.read_int(2)
        fragmented_free_bytes = self.read_int(1)

        cells_pos = [self.read_int(2) for _ in range(cells_num)]
        rows = []

        for cell_pos in cells_pos:
            self.pos = self.page_size * page_num + cell_pos

            payload_len = self.read_varint()
            rowid = self.read_varint()

            columns_type = []

            payload_pos = self.pos
            header_size = self.read_varint()
            while self.pos < payload_pos + header_size:
                column_type = self.read_varint()
                columns_type.append(column_type)

            cells = []

            for column_type in columns_type:
                if column_type == 0:
                    data = rowid
                elif 1 <= column_type <= 4:
                    data = self.read_int(column_type)
                elif column_type == 5:
                    data = self.read_int(6)
                elif column_type == 6:
                    data = self.read_int(8)
                elif column_type == 7:
                    # TODO: float
                    data = self.read(8)
                elif column_type == 8:
                    data = 0
                elif column_type == 9:
                    data = 1
                elif column_type >= 12 and column_type % 2 == 0:
                    length = int((column_type - 12) / 2)
                    data = self.read(length)
                else:
                    length = int((column_type - 13) / 2)
                    data = self.read(length).decode()

                cells.append(data)

            rows.append(cells)

        return rows

    def _read_interior_table(self, page_num: int):
        first_block = self.read_int(2)
        cells_num = self.read_int(2)
        cells_pos = self.read_int(2)
        fragmented_free_bytes = self.read_int(1)
        last_page_num = self.read_int(4)

        cells_pos = [self.read_int(2) for _ in range(cells_num)]
        rows = []

        for cell_pos in cells_pos:
            self.pos = self.page_size * page_num + cell_pos
            child_page_num = self.read_int(4)
            rowid = self.read_varint()
            rows += self.read_page(child_page_num - 1)

        return rows + self.read_page(last_page_num - 1)

    def read_table(self, name: str):
        page = next(t[3] - 1 for t in self.tables if t[1] == name)
        return self.read_page(page)



#!/usr/bin/env python3
"""Support for Tuya Cover."""

import logging
from typing import Any, List

from homeassistant.components.cover import (
    DEVICE_CLASS_CURTAIN,
    DOMAIN as DEVICE_DOMAIN,
    SUPPORT_CLOSE,
    SUPPORT_OPEN,
    SUPPORT_SET_POSITION,
    SUPPORT_STOP,
    CoverEntity,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import DOMAIN, TUYA_DEVICE_MANAGER, TUYA_DISCOVERY_NEW, TUYA_HA_TUYA_MAP

_LOGGER = logging.getLogger(__name__)

TUYA_SUPPORT_TYPE = {"cl", "clkg"}  # Curtain  # Curtain Switch

# Curtain
# https://developer.tuya.com/en/docs/iot/f?id=K9gf46o5mtfyc
DPCODE_CONTROL = "control"
DPCODE_PERCENT_CONTROL = "percent_control"

ATTR_POSITION = "position"


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya cover dynamically through tuya discovery."""
    print("cover init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya cover."""
        print("cover add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: List):
    """Set up Tuya Cover."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue
        entities.append(TuyaHaCover(device, device_manager))
    return entities


class TuyaHaCover(TuyaHaDevice, CoverEntity):
    """Tuya Switch Device."""

    # property
    @property
    def device_class(self) -> str:
        """Return Entity Properties."""
        return DEVICE_CLASS_CURTAIN

    @property
    def current_cover_position(self) -> int:
        """Return cover current position."""
        return self.tuya_device.status.get(DPCODE_PERCENT_CONTROL, 0)

    def open_cover(self, **kwargs: Any) -> None:
        """Open the cover."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_CONTROL, "value": "open"}]
        )

    def close_cover(self, **kwargs: Any) -> None:
        """Close cover."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_CONTROL, "value": "close"}]
        )

    def stop_cover(self, **kwargs):
        """Stop the cover."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_CONTROL, "value": "stop"}]
        )

    def set_cover_position(self, **kwargs):
        """Move the cover to a specific position."""
        print("cover-->", kwargs)
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id,
            [{"code": DPCODE_PERCENT_CONTROL, "value": kwargs[ATTR_POSITION]}],
        )

    @property
    def supported_features(self):
        """Flag supported features."""
        supports = SUPPORT_OPEN | SUPPORT_CLOSE | SUPPORT_STOP

        if DPCODE_PERCENT_CONTROL in self.tuya_device.status:
            supports = supports | SUPPORT_SET_POSITION

        return supports



#!/usr/bin/env python3
"""Config flow for Tuya."""

import logging

from tuya_iot import ProjectType, TuyaOpenAPI
import voluptuous as vol

from homeassistant import config_entries

from .const import (
    CONF_ACCESS_ID,
    CONF_ACCESS_SECRET,
    CONF_APP_TYPE,
    CONF_COUNTRY_CODE,
    CONF_ENDPOINT,
    CONF_PASSWORD,
    CONF_PROJECT_TYPE,
    CONF_USERNAME,
    DOMAIN,
    TUYA_APP_TYPE,
    TUYA_ENDPOINT,
    TUYA_PROJECT_TYPE,
)

RESULT_SINGLE_INSTANCE = "single_instance_allowed"

_LOGGER = logging.getLogger(__name__)
RESULT_AUTH_FAILED = "invalid_auth"

# Project Type
DATA_SCHEMA_PROJECT_TYPE = vol.Schema(
    {vol.Required(CONF_PROJECT_TYPE, default=0): vol.In(TUYA_PROJECT_TYPE)}
)

# INDUSTRY_SOLUTIONS Schema
DATA_SCHEMA_INDUSTRY_SOLUTIONS = vol.Schema(
    {
        vol.Required(CONF_ENDPOINT): vol.In(TUYA_ENDPOINT),
        vol.Required(CONF_ACCESS_ID): str,
        vol.Required(CONF_ACCESS_SECRET): str,
        vol.Required(CONF_USERNAME): str,
        vol.Required(CONF_PASSWORD): str,
    }
)

# SMART_HOME Schema
DATA_SCHEMA_SMART_HOME = vol.Schema(
    {
        vol.Required(CONF_ENDPOINT): vol.In(TUYA_ENDPOINT),
        vol.Required(CONF_ACCESS_ID): str,
        vol.Required(CONF_ACCESS_SECRET): str,
        vol.Required(CONF_APP_TYPE): vol.In(TUYA_APP_TYPE),
        vol.Required(CONF_COUNTRY_CODE): str,
        vol.Required(CONF_USERNAME): str,
        vol.Required(CONF_PASSWORD): str,
    }
)


class TuyaConfigFlow(config_entries.ConfigFlow, domain=DOMAIN):
    """Tuya Config Flow."""

    conf_project_type = None

    def _try_login(self, user_input):
        print("TuyaConfigFlow._try_login start, user_input:", user_input)
        project_type = ProjectType(user_input[CONF_PROJECT_TYPE])
        api = TuyaOpenAPI(
            user_input[CONF_ENDPOINT],
            user_input[CONF_ACCESS_ID],
            user_input[CONF_ACCESS_SECRET],
            project_type,
        )
        api.set_dev_channel("hass")

        response = (
            api.login(user_input[CONF_USERNAME], user_input[CONF_PASSWORD])
            if project_type == ProjectType.INDUSTY_SOLUTIONS
            else api.login(
                user_input[CONF_USERNAME],
                user_input[CONF_PASSWORD],
                user_input[CONF_COUNTRY_CODE],
                user_input[CONF_APP_TYPE],
            )
        )

        print("TuyaConfigFlow._try_login finish, response:", response)
        return response

    async def async_step_import(self, user_input=None):
        """Step import."""
        return await self.async_step_user(user_input, is_import=True)

    async def async_step_project_type(self, user_input=None):
        """Step project type."""
        self.conf_project_type = user_input[CONF_PROJECT_TYPE]
        self.project_type = ProjectType(self.conf_project_type)
        return (
            self.async_show_form(step_id="user", data_schema=DATA_SCHEMA_SMART_HOME)
            if self.project_type == ProjectType.SMART_HOME
            else self.async_show_form(
                step_id="user", data_schema=DATA_SCHEMA_INDUSTRY_SOLUTIONS
            )
        )

    async def async_step_user(self, user_input=None, is_import=False):
        """Step user."""
        print("TuyaConfigFlow.async_step_user start, is_import=", user_input)

        if self._async_current_entries():
            return self.async_abort(reason=RESULT_SINGLE_INSTANCE)

        errors = {}
        if user_input is not None:
            if self.conf_project_type is not None:
                user_input[CONF_PROJECT_TYPE] = self.conf_project_type

            response = await self.hass.async_add_executor_job(
                self._try_login, user_input
            )

            if response.get("success", False):
                print("TuyaConfigFlow.async_step_user login success")
                return self.async_create_entry(
                    title=user_input[CONF_USERNAME],
                    data=user_input,
                )
            else:
                errors["base"] = RESULT_AUTH_FAILED
                if is_import:
                    return self.async_abort(reason=errors["base"])
                else:
                    return (
                        self.async_show_form(
                            step_id="user",
                            data_schema=DATA_SCHEMA_SMART_HOME,
                            errors=errors,
                        )
                        if self.project_type == ProjectType.SMART_HOME
                        else self.async_show_form(
                            step_id="user",
                            data_schema=DATA_SCHEMA_INDUSTRY_SOLUTIONS,
                            errors=errors,
                        )
                    )

        return self.async_show_form(
            step_id="project_type", data_schema=DATA_SCHEMA_PROJECT_TYPE, errors=errors
        )



#!/usr/bin/env python3
"""Support for Tuya Smart devices."""

from functools import partial
import functools
from ..utils import utils
from .. import fhem
from ..generic import FhemModule

import itertools
from typing import Any

from tuya_iot import (
    ProjectType,
    TuyaDevice,
    TuyaDeviceListener,
    TuyaDeviceManager,
    TuyaHomeManager,
    TuyaOpenAPI,
    TuyaOpenMQ,
)

from .const import (
    CONF_ACCESS_ID,
    CONF_ACCESS_SECRET,
    CONF_APP_TYPE,
    CONF_COUNTRY_CODE,
    CONF_ENDPOINT,
    CONF_PASSWORD,
    CONF_PROJECT_TYPE,
    CONF_USERNAME,
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
    TUYA_MQTT_LISTENER,
    TUYA_SUPPORT_HA_TYPE,
    TUYA_SETUP_PLATFORM,
)


class tuya_cloud(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    async def Define(self, hash, args, argsh):
        project_type = ProjectType(argsh[CONF_PROJECT_TYPE])
        api = TuyaOpenAPI(
            argsh[CONF_ENDPOINT],
            argsh[CONF_ACCESS_ID],
            argsh[CONF_ACCESS_SECRET],
            project_type,
        )

        api.set_dev_channel("hass")

        response = (
            await utils.run_blocking(
                functools.partial(api.login, argsh[CONF_USERNAME], argsh[CONF_PASSWORD])
            )
            if project_type == ProjectType.INDUSTY_SOLUTIONS
            else await utils.run_blocking(
                functools.partial(
                    api.login,
                    argsh[CONF_USERNAME],
                    argsh[CONF_PASSWORD],
                    argsh[CONF_COUNTRY_CODE],
                    argsh[CONF_APP_TYPE],
                )
            )
        )
        if response.get("success", False) is False:
            self.logger.error(
                "Tuya login error response: %s",
                response,
            )
            return False

        mq = TuyaOpenMQ(api)
        mq.start()

        device_manager = TuyaDeviceManager(api, mq)

        # Get device list
        home_manager = TuyaHomeManager(api, mq, device_manager)
        await utils.run_blocking(functools.partial(home_manager.updateDeviceCache))

        class DeviceListener(TuyaDeviceListener):
            def updateDevice(self, device: TuyaDevice):
                # TODO
                # get all tuya devices from FHEM
                # send update state to device with device.id
                for haDevice in hass.data[DOMAIN][TUYA_HA_DEVICES]:
                    if haDevice.tuya_device.id == device.id:
                        print("_update-->", self, ";->>", haDevice.tuya_device.status)
                        haDevice.schedule_update_ha_state()

            def addDevice(self, device: TuyaDevice):
                print("tuya device add-->", device)

                device_add = False

                print(
                    "add device category->",
                    device.category,
                    "; keys->",
                    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].keys(),
                )
                if device.category in itertools.chain(
                    *hass.data[DOMAIN][TUYA_HA_TUYA_MAP].values()
                ):
                    map = hass.data[DOMAIN][TUYA_HA_TUYA_MAP]

                    remove_device(hass, device.id)

                    for key, list in map.items():
                        if device.category in list:
                            device_add = True
                            async_dispatcher_send(
                                hass, TUYA_DISCOVERY_NEW.format(key), [device.id]
                            )

                if device_add:
                    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
                    device_manager.mq.stop()
                    mq = TuyaOpenMQ(device_manager.api)
                    mq.start()

                    device_manager.mq = mq
                    mq.add_message_listener(device_manager._onMessage)

            def removeDevice(self, id: str):
                print("tuya remove device:", id)
                remove_device(hass, id)

        __listener = DeviceListener()
        hass.data[DOMAIN][TUYA_MQTT_LISTENER] = __listener
        device_manager.addDeviceListener(__listener)
        hass.data[DOMAIN][TUYA_DEVICE_MANAGER] = device_manager

        # Clean up device entities
        await cleanup_device_registry(hass)

        print("domain key->", str(hass.data[DOMAIN][TUYA_HA_TUYA_MAP]))
        print("init support type->", TUYA_SUPPORT_HA_TYPE)

        for platform in TUYA_SUPPORT_HA_TYPE:
            print("tuya async platform-->", platform)
            hass.async_create_task(
                hass.config_entries.async_forward_entry_setup(entry, platform)
            )
            hass.data[DOMAIN][TUYA_SETUP_PLATFORM].add(platform)

        return True


async def cleanup_device_registry(hass: HomeAssistant):
    """Remove deleted device registry entry if there are no remaining entities."""

    device_registry = hass.helpers.device_registry.async_get(hass)
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]

    for dev_id, device_entity in list(device_registry.devices.items()):
        for item in device_entity.identifiers:
            if DOMAIN == item[0] and item[1] not in device_manager.deviceMap.keys():
                device_registry.async_remove_device(dev_id)
                break


def remove_device(hass: HomeAssistant, device_id: str):
    """Remove device from hass cache."""
    device_registry = hass.helpers.device_registry.async_get(hass)
    entity_registry = hass.helpers.entity_registry.async_get(hass)
    for entity in list(entity_registry.entities.values()):
        if entity.unique_id.startswith(f"ty{device_id}"):
            entity_registry.async_remove(entity.entity_id)
            if device_registry.async_get(entity.device_id):
                device_registry.async_remove_device(entity.device_id)


async def async_setup(hass, config):
    """Set up the Tuya integration."""

    conf = config.get(DOMAIN)

    print("Tuya async setup conf %s \n" % conf)
    if conf is not None:

        async def flow_init() -> Any:
            try:
                result = await hass.config_entries.flow.async_init(
                    DOMAIN, context={"source": SOURCE_IMPORT}, data=conf
                )
            except Exception as inst:
                print(inst.args)
            print("Tuya async setup flow_init")
            return result

        hass.async_create_task(flow_init())

    return True


async def async_unload_entry(hass: HomeAssistant, entry: ConfigEntry):
    """Unloading the Tuya platforms."""
    print("integration unload")
    unload = await hass.config_entries.async_unload_platforms(
        entry, hass.data[DOMAIN]["setup_platform"]
    )
    if unload:
        __device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
        __device_manager.mq.stop()
        __device_manager.removeDeviceListener(hass.data[DOMAIN][TUYA_MQTT_LISTENER])

        hass.data.pop(DOMAIN)

    return unload


async def async_setup_entry(hass: HomeAssistant, entry: ConfigEntry):
    """Async setup hass config entry."""
    print("tuya.__init__.async_setup_entry-->", entry.data)

    hass.data[DOMAIN] = {TUYA_HA_TUYA_MAP: {}, TUYA_HA_DEVICES: []}
    hass.data[DOMAIN][TUYA_SETUP_PLATFORM] = set()

    success = await _init_tuya_sdk(hass, entry)
    if not success:
        return False

    return True



#!/usr/bin/env python3
"""Support for Tuya switches."""
from __future__ import annotations

import logging
from typing import Any

from tuya_iot import TuyaDevice, TuyaDeviceManager

from homeassistant.components.switch import DOMAIN as DEVICE_DOMAIN, SwitchEntity
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)

TUYA_SUPPORT_TYPE = {
    "kg",  # Switch
    "cz",  # Socket
    "pc",  # Power Strip
}

# Switch(kg), Socket(cz), Power Strip(pc)
# https://developer.tuya.com/docs/iot/open-api/standard-function/electrician-category/categorykgczpc?categoryId=486118
DPCODE_SWITCH = "switch"


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya sensors dynamically through tuya discovery."""
    print("switch init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya sensor."""
        print("switch add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: list):
    """Set up Tuya Switch device."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue

        switch_set = []
        for function in device.function:
            if function.startswith(DPCODE_SWITCH):
                switch_set.append(function.replace(DPCODE_SWITCH, ""))

        if len(switch_set) > 1:
            for channel in switch_set:
                entities.append(TuyaHaSwitch(device, device_manager, channel))
        elif len(switch_set) == 1:
            entities.append(TuyaHaSwitch(device, device_manager))

    return entities


class TuyaHaSwitch(TuyaHaDevice, SwitchEntity):
    """Tuya Switch Device."""

    dp_code_switch = DPCODE_SWITCH

    def __init__(
        self, device: TuyaDevice, deviceManager: TuyaDeviceManager, channel: str = ""
    ) -> None:
        """Init TuyaHaSwitch."""
        super().__init__(device, deviceManager)

        self.channel = channel
        if len(channel) > 0:
            self.dp_code_switch = DPCODE_SWITCH + self.channel
        else:
            for function in device.function:
                if function.startswith(DPCODE_SWITCH):
                    self.dp_code_switch = function

    @property
    def unique_id(self) -> str | None:
        """Return a unique ID."""
        return f"{super().unique_id}{self.channel}"

    @property
    def name(self) -> str | None:
        """Return Tuya device name."""
        return self.tuya_device.name + self.channel

    @property
    def is_on(self) -> bool:
        """Return true if switch is on."""
        return self.tuya_device.status.get(self.dp_code_switch, False)

    def turn_on(self, **kwargs: Any) -> None:
        """Turn the switch on."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": self.dp_code_switch, "value": True}]
        )

    def turn_off(self, **kwargs: Any) -> None:
        """Turn the device off."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": self.dp_code_switch, "value": False}]
        )



#!/usr/bin/env python3
"""Support for Tuya switches."""

import json
import logging
from typing import List, Optional, Tuple

from tuya_iot import TuyaDevice, TuyaDeviceManager

from homeassistant.components.number import DOMAIN as DEVICE_DOMAIN, NumberEntity
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)

TUYA_SUPPORT_TYPE = {
    "hps",  # Human Presence Sensor
}

# Switch(kg), Socket(cz), Power Strip(pc)
# https://developer.tuya.com/docs/iot/open-api/standard-function/electrician-category/categorykgczpc?categoryId=486118
DPCODE_SENSITIVITY = "sensitivity"


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya number dynamically through tuya discovery."""
    print("switch init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya number."""
        print("switch add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: List):
    """Set up Tuya Switch device."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue

        if DPCODE_SENSITIVITY in device.status:
            entities.append(TuyaHaNumber(device, device_manager, DPCODE_SENSITIVITY))

    return entities


class TuyaHaNumber(TuyaHaDevice, NumberEntity):
    """Tuya Device Number."""

    def __init__(
        self, device: TuyaDevice, deviceManager: TuyaDeviceManager, code: str = ""
    ):
        """Init tuya number device."""
        self._code = code
        super().__init__(device, deviceManager)

    # ToggleEntity

    def set_value(self, value: float) -> None:
        """Update the current value."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": self._code, "value": int(value)}]
        )

    @property
    def unique_id(self) -> Optional[str]:
        """Return a unique ID."""
        return f"{super().unique_id}{self._code}"

    @property
    def value(self) -> float:
        """Return current value."""
        return self.tuya_device.status.get(self._code, 0)

    @property
    def min_value(self) -> float:
        """Return min value."""
        return self._get_code_range()[0]

    @property
    def max_value(self) -> float:
        """Return max value."""
        return self._get_code_range()[1]

    @property
    def step(self) -> float:
        """Return step."""
        return self._get_code_range()[2]

    def _get_code_range(self) -> Tuple[int, int, int]:
        range = json.loads(self.tuya_device.function.get(self._code).values)
        return range.get("min", 0), range.get("max", 0), range.get("step", 0)



#!/usr/bin/env python3
"""Support for Tuya Humidifiers."""

import json
import logging
from typing import List

from homeassistant.components.humidifier import (
    DOMAIN as DEVICE_DOMAIN,
    SUPPORT_MODES,
    HumidifierEntity,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)

TUYA_SUPPORT_TYPE = {
    "jsq",  # Humidifier
}

# Humidifier(jsq)
# https://developers.home-assistant.io/docs/core/entity/humidifier
DPCODE_MODE = "mode"
DPCODE_SWITCH = "switch"


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya sensors dynamically through tuya discovery."""
    print("humidifier init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya sensor."""
        print("humidifier add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: List):
    """Set up Tuya Switch device."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue

        entities.append(TuyaHaHumidifier(device, device_manager))

    return entities


class TuyaHaHumidifier(TuyaHaDevice, HumidifierEntity):
    """Tuya Switch Device."""

    @property
    def is_on(self):
        """Return the device is on or off."""
        return self.tuya_device.status.get(DPCODE_SWITCH, False)

    @property
    def mode(self):
        """Return the current mode."""
        return self.tuya_device.status.get(DPCODE_MODE)

    @property
    def available_modes(self):
        """Return a list of available modes."""
        return json.loads(self.tuya_device.function.get(DPCODE_MODE, {}).values).get(
            "range"
        )

    @property
    def supported_features(self):
        """Return humidifier support features."""
        supports = 0
        if DPCODE_MODE in self.tuya_device.status:
            supports = supports | SUPPORT_MODES
        return supports

    def set_mode(self, mode):
        """Set new target preset mode."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_MODE, "value": mode}]
        )

    def turn_on(self, **kwargs):
        """Turn the device on."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_SWITCH, "value": True}]
        )

    def turn_off(self, **kwargs):
        """Turn the device off."""
        self.tuya_device_manager.sendCommands(
            self.tuya_device.id, [{"code": DPCODE_SWITCH, "value": False}]
        )



#!/usr/bin/env python3
"""Support for Tuya switches."""

import logging
import json
from typing import List, Optional

from tuya_iot import TuyaDevice, TuyaDeviceManager

from .base import TuyaHaDevice
from .const import (
    DOMAIN,
    TUYA_DEVICE_MANAGER,
    TUYA_DISCOVERY_NEW,
    TUYA_HA_DEVICES,
    TUYA_HA_TUYA_MAP,
)

_LOGGER = logging.getLogger(__name__)

TUYA_SUPPORT_TYPE = [
    "wsdcg",  # Temperature and Humidity Sensor
    "mcs",  # Door Window Sensor
    "ywbj",  # Somke Detector
    "rqbj",  # Gas Detector
    "pir",  # PIR Detector
    "sj",  # Water Detector
    "pm2.5",  # PM2.5 Sensor
]

# Smoke Detector
# https://developer.tuya.com/en/docs/iot/s?id=K9gf48r5i2iiy
DPCODE_BATTERY = "va_battery"
DPCODE_BATTERY_PERCENTAGE = "battery_percentage"
DPCODE_BATTERY_CODE = "battery"

DPCODE_TEMPERATURE = "va_temperature"
DPCODE_HUMIDITY = "va_humidity"

DPCODE_PM100_VALUE = "pm100_value"
DPCODE_PM25_VALUE = "pm25_value"
DPCODE_PM10_VALUE = "pm10_value"

DPCODE_TEMP_CURRENT = "temp_current"
DPCODE_HUMIDITY_VALUE = "humidity_value"


async def async_setup_entry(
    hass: HomeAssistant, entry: ConfigEntry, async_add_entities
):
    """Set up tuya sensors dynamically through tuya discovery."""
    print("sensor init")

    hass.data[DOMAIN][TUYA_HA_TUYA_MAP].update({DEVICE_DOMAIN: TUYA_SUPPORT_TYPE})

    async def async_discover_device(dev_ids):
        """Discover and add a discovered tuya sensor."""
        print("sensor add->", dev_ids)
        if not dev_ids:
            return
        entities = await hass.async_add_executor_job(_setup_entities, hass, dev_ids)
        hass.data[DOMAIN][TUYA_HA_DEVICES].extend(entities)
        async_add_entities(entities)

    async_dispatcher_connect(
        hass, TUYA_DISCOVERY_NEW.format(DEVICE_DOMAIN), async_discover_device
    )

    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    device_ids = []
    for (device_id, device) in device_manager.deviceMap.items():
        if device.category in TUYA_SUPPORT_TYPE:
            device_ids.append(device_id)
    await async_discover_device(device_ids)


def _setup_entities(hass, device_ids: List):
    """Set up Tuya Switch device."""
    device_manager = hass.data[DOMAIN][TUYA_DEVICE_MANAGER]
    entities = []
    for device_id in device_ids:
        device = device_manager.deviceMap[device_id]
        if device is None:
            continue

        if DPCODE_BATTERY in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_BATTERY,
                    DPCODE_BATTERY,
                    PERCENTAGE,
                )
            )
        if DPCODE_BATTERY_PERCENTAGE in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_BATTERY,
                    DPCODE_BATTERY_PERCENTAGE,
                    PERCENTAGE,
                )
            )
        if DPCODE_BATTERY_CODE in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_BATTERY,
                    DPCODE_BATTERY_CODE,
                    PERCENTAGE,
                )
            )

        if DPCODE_TEMPERATURE in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_TEMPERATURE,
                    DPCODE_TEMPERATURE,
                    TEMP_CELSIUS,
                )
            )
        if DPCODE_TEMP_CURRENT in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_TEMPERATURE,
                    DPCODE_TEMP_CURRENT,
                    TEMP_CELSIUS,
                )
            )

        if DPCODE_HUMIDITY in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_HUMIDITY,
                    DPCODE_HUMIDITY,
                    PERCENTAGE,
                )
            )
        if DPCODE_HUMIDITY_VALUE in device.status:
            entities.append(
                TuyaHaSensor(
                    device,
                    device_manager,
                    DEVICE_CLASS_HUMIDITY,
                    DPCODE_HUMIDITY_VALUE,
                    PERCENTAGE,
                )
            )

        if DPCODE_PM100_VALUE in device.status:
            entities.append(
                TuyaHaSensor(
                    device, device_manager, "PM10", DPCODE_PM100_VALUE, "ug/m³"
                )
            )
        if DPCODE_PM25_VALUE in device.status:
            entities.append(
                TuyaHaSensor(
                    device, device_manager, "PM2.5", DPCODE_PM25_VALUE, "ug/m³"
                )
            )
        if DPCODE_PM10_VALUE in device.status:
            entities.append(
                TuyaHaSensor(
                    device, device_manager, "PM1.0", DPCODE_PM10_VALUE, "ug/m³"
                )
            )

    return entities


class TuyaHaSensor(TuyaHaDevice, SensorEntity):
    """Tuya Sensor Device."""

    def __init__(
        self,
        device: TuyaDevice,
        deviceManager: TuyaDeviceManager,
        sensor_type: str,
        sensor_code: str,
        sensor_unit: str,
    ):
        """Init TuyaHaSensor."""
        self._type = sensor_type
        self._code = sensor_code
        self._unit = sensor_unit
        super().__init__(device, deviceManager)

    @property
    def unique_id(self) -> Optional[str]:
        """Return a unique ID."""
        return f"{super().unique_id}{self._code}"

    @property
    def name(self):
        """Return the name of the sensor."""
        return self.tuya_device.name + "_" + self._type

    @property
    def state(self):
        """Return the state of the sensor."""
        __value = self.tuya_device.status.get(self._code)
        print(self.tuya_device.status_range.get(self._code).values)
        if self.tuya_device.status_range.get(self._code).type == "Integer":
            __value_range = json.loads(
                self.tuya_device.status_range.get(self._code).values
            )
            __state = (__value) * 1.0 / (10 ** __value_range.get("scale"))
            if __value_range.get("scale") == 0:
                return int(__state)
            else:
                return f"%.{__value_range.get('scale')}f" % __state
        return ""

    @property
    def unit_of_measurement(self):
        """Return the units of measurement."""
        return self._unit

    @property
    def device_class(self):
        """Device class of this entity."""
        return self._type

    @property
    def available(self) -> bool:
        """Return if the device is available."""
        return True



import json
import logging
import random
import re
import socket
import string
import time
from threading import Thread
from typing import Optional

from paho.mqtt.client import Client, MQTTMessage
from . import bluetooth, utils
from .mini_miio import SyncmiIO
from .shell import TelnetShell, ntp_time
from .unqlite import Unqlite, SQLite
from .utils import GLOBAL_PROP

_LOGGER = logging.getLogger(__name__)

RE_NWK_KEY = re.compile(r"lumi send-nwk-key (0x.+?) {(.+?)}")
RE_MAC = re.compile("^0x0*")
# MAC reverse
RE_REVERSE = re.compile(r"(..)(..)(..)(..)(..)(..)")


class GatewayV:
    """Handling different firmware versions."""

    ver = None

    @property
    def ver_mesh_group(self) -> str:
        return "mesh_group_v1" if self.ver >= "1.4.6_0043" else "mesh_group"

    @property
    def ver_zigbee_db(self) -> str:
        # https://github.com/AlexxIT/XiaomiGateway3/issues/14
        # fw 1.4.6_0012 and below have one zigbee_gw.db file
        # fw 1.4.6_0030 have many json files in this folder
        return "*.json" if self.ver >= "1.4.6_0030" else "zigbee_gw.db"

    @property
    def ver_miio(self) -> bool:
        return self.ver >= "1.4.7_0063"

    @property
    def ver_z3(self) -> bool:
        return self.ver >= "1.4.7_0063"


class GatewayMesh:
    enabled: bool = None

    devices: dict = None
    updates: dict = None

    miio: SyncmiIO = None
    mesh_params: list = None
    mesh_ts: float = 0

    def debug(self, message: str):
        raise NotImplemented

    def mesh_start(self):
        self.mesh_params = []

        for device in self.devices.values():
            if device["type"] == "mesh" and "childs" not in device:
                # TODO: rewrite more clear logic for lights and switches
                p = device["params"][0]
                self.mesh_params.append(
                    {"did": device["did"], "siid": p[0], "piid": p[1]}
                )

        if self.mesh_params:
            self.mesh_ts = time.time() + 30
            Thread(target=self.mesh_run).start()

    def mesh_run(self):
        self.debug("Start Mesh Thread")

        while self.enabled:
            if time.time() < self.mesh_ts:
                time.sleep(1)
                continue

            try:
                resp = self.miio.send_bulk("get_properties", self.mesh_params)
                if resp:
                    params2 = []
                    for item in resp:
                        if "value" not in item:
                            continue

                        did = item["did"]
                        device_params = self.devices[did]["params"]
                        # get other props for turn on lights or live switches
                        if device_params[0][3] == "switch" or item["value"]:
                            params2 += [
                                {"did": did, "siid": p[0], "piid": p[1]}
                                for p in device_params[1:]
                            ]

                    if params2:
                        resp2 = self.miio.send_bulk("get_properties", params2)
                        if resp2:
                            resp += resp2

                    self.debug(f"Pull Mesh {resp}")
                    self.process_mesh_data(resp)

                else:
                    self.debug("Can't get mesh bulb state")

            except Exception as e:
                self.debug(f"ERROR in mesh thread {e}")

            self.mesh_ts = time.time() + 30

    def process_mesh_data(self, data: list):
        """Can receive multiple properties from multiple devices.

        data = [{'did':123,'siid':2,'piid':1,'value:True}]
        """
        bulk = {}

        for param in data:
            if param.get("code", 0) != 0:
                continue

            did = param["did"]
            if did not in self.updates:
                continue

            device = self.devices[did]

            prop = next(
                (
                    p[2]
                    for p in device["params"]
                    if p[0] == param["siid"] and p[1] == param["piid"]
                ),
                None,
            )
            if not prop:
                continue

            if did not in bulk:
                bulk[did] = {}

            bulk[did][prop] = param["value"]

        for did, payload in bulk.items():
            self.debug(f"Process Mesh Data for {did}: {payload}")
            for handler in self.updates[did]:
                handler(payload)

    def send_mesh(self, device: dict, data: dict):
        # data = {'light':True}
        payload = []
        for k, v in data.items():
            param = next(p for p in device["params"] if p[2] == k)
            payload.append(
                {
                    "did": device["did"],
                    "siid": param[0],
                    "piid": param[1],
                    "value": v if param[0] != 8 else int(v),
                }
            )

        try:
            # 2 seconds are selected experimentally
            if self.miio.send("set_properties", payload):
                self.mesh_force_update()
        except Exception:
            self.debug(f"Can't send mesh {device['did']} => {data}")

    def mesh_force_update(self):
        self.mesh_ts = time.time() + 2


# noinspection PyUnusedLocal
class GatewayStats:
    stats: dict = None
    host: str = None
    info_ts: float = 0

    mqtt: Client = None
    gw_topic: str = None

    # if mqtt connected
    available: bool = None

    # interval for auto parent refresh in minutes, 0 - disabled auto refresh
    # -1 - disabled
    parent_scan_interval: Optional[int] = None

    # collected data from MQTT topic log/z3 (zigbee console)
    z3buffer: Optional[dict] = None

    def debug(self, message: str):
        raise NotImplemented

    def add_stats(self, ieee: str, handler):
        self.stats[ieee] = handler

        if self.parent_scan_interval > 0:
            self.info_ts = time.time() + 5

    def remove_stats(self, ieee: str, handler):
        self.stats.pop(ieee)

    def process_gw_stats(self, payload: dict = None):
        # empty payload - update available state
        self.debug(f"gateway <= {payload or self.available}")

        if "lumi.0" not in self.stats:
            return

        if payload:
            if "networkUp" in payload:
                # {"networkUp":false}
                payload = {
                    "network_pan_id": payload.get("networkPanId"),
                    "radio_tx_power": payload.get("radioTxPower"),
                    "radio_channel": payload.get("radioChannel"),
                }
            elif "free_mem" in payload:
                s = payload["run_time"]
                h, m, s = s // 3600, s % 3600 // 60, s % 60
                payload = {
                    "free_mem": payload["free_mem"],
                    "load_avg": payload["load_avg"],
                    "rssi": -payload["rssi"],
                    "uptime": f"{h:02}:{m:02}:{s:02}",
                }

        self.stats["lumi.0"](payload)

    def process_zb_stats(self, payload: dict):
        ieee = payload["eui64"]
        if ieee in self.stats:
            self.stats[ieee](payload)

        if self.info_ts and time.time() > self.info_ts:
            # block any auto updates in 30 seconds
            self.info_ts = time.time() + 30

            self.get_gateway_info()

    def process_ble_stats(self, payload: dict):
        did = payload["dev"]["did"]
        if did in self.stats:
            self.stats[did](payload)

    def process_z3(self, payload: str):
        if payload.startswith("CLI command executed"):
            cmd = payload[22:-1]
            if cmd == "debugprint all_on" or self.z3buffer is None:
                # reset all buffers
                self.z3buffer = {}
            else:
                self.z3buffer[cmd] = self.z3buffer["buffer"]

            self.z3buffer["buffer"] = ""

            if cmd == "plugin concentrator print-table":
                self._process_gateway_info()

        elif self.z3buffer:
            self.z3buffer["buffer"] += payload

    def get_gateway_info(self):
        self.debug("Update zigbee network info")
        payload = {
            "commands": [
                {"commandcli": "debugprint all_on"},
                {"commandcli": "plugin device-table print"},
                {"commandcli": "plugin stack-diagnostics child-table"},
                {"commandcli": "plugin stack-diagnostics neighbor-table"},
                {"commandcli": "plugin concentrator print-table"},
                {"commandcli": "debugprint all_off"},
            ]
        }
        payload = json.dumps(payload, separators=(",", ":"))
        self.mqtt.publish(self.gw_topic + "commands", payload)

    def send_zigbee_cli(self, commands: list):
        payload = {"commands": [{"commandcli": cmd} for cmd in commands]}
        payload = json.dumps(payload, separators=(",", ":"))
        self.mqtt.publish(self.gw_topic + "commands", payload)

    def _process_gateway_info(self):
        self.debug("Update parent info table")

        try:
            raw = self.z3buffer["plugin device-table print"]
            m1 = re.findall(
                r"\d+ ([A-F0-9]{4}): {2}([A-F0-9]{16}) 0 {2}\w+ " r"(\d+)", raw
            )

            raw = self.z3buffer["plugin stack-diagnostics child-table"]
            m2 = re.findall(r"\(>\)([A-F0-9]{16})", raw)

            raw = self.z3buffer["plugin stack-diagnostics neighbor-table"]
            m3 = re.findall(r"\(>\)([A-F0-9]{16})", raw)

            raw = self.z3buffer["plugin concentrator print-table"]
            m4 = re.findall(r": ([A-F0-9x> -]{16,}) -> 0x0000", raw)
            m4 = [i.replace("0x", "").split(" -> ") for i in m4]
            m4 = {i[0]: i[1:] for i in m4}

            self.debug(f"Total zigbee devices: {len(m1)}")

            for i in m1:
                ieee = "0x" + i[1]

                nwk = i[0]
                ago = int(i[2])
                type_ = "device" if i[1] in m2 else "router" if i[1] in m3 else "-"
                parent = "0x" + m4[nwk][0] if nwk in m4 else "-"

                payload = {
                    "nwk": "0x" + nwk,
                    "ago": ago,
                    "type": type_,
                    "parent": parent,
                }

                if ieee in self.stats:
                    self.stats[ieee](payload)
                else:
                    self.debug(f"Unknown zigbee device {ieee}: {payload}")

            # one hour later
            if self.parent_scan_interval > 0:
                self.info_ts = time.time() + self.parent_scan_interval * 60

        except Exception as e:
            self.debug(f"Can't update parents: {e}")


# noinspection PyUnusedLocal
class Gateway3(Thread, GatewayV, GatewayMesh, GatewayStats):
    time_offset = 0
    pair_model = None
    pair_payload = None

    def __init__(self, host: str, token: str, config: dict, **options):
        super().__init__(daemon=True)

        self.host = host
        self.options = options

        self.miio = SyncmiIO(host, token)

        self.mqtt = Client()
        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect
        self.mqtt.on_message = self.on_message

        self._ble = options.get("ble")  # for fast access
        self._debug = options.get("debug", "")  # for fast access
        self.parent_scan_interval = (
            -1 if options.get("parent") is None else options["parent"]
        )
        self.default_devices = config["devices"] if config else None

        if "true" in self._debug:
            self.miio.debug = True

        self.devices = {}
        self.updates = {}
        self.setups = {}
        self.stats = {}

    @property
    def device(self):
        return self.devices["lumi.0"]

    def add_update(self, did: str, handler):
        """Add handler to device update event."""
        self.updates.setdefault(did, []).append(handler)

    def remove_update(self, did: str, handler):
        self.updates.setdefault(did, []).remove(handler)

    def add_setup(self, domain: str, handler):
        """Add hass device setup funcion."""
        self.setups[domain] = handler

    def debug(self, message: str):
        # basic logs
        if "true" in self._debug:
            _LOGGER.debug(f"{self.host} | {message}")

    def stop(self):
        self.enabled = False
        self.mqtt._thread_terminate = True

    def run(self):
        """Main thread loop."""
        self.debug("Start main thread")

        self.mqtt.connect_async(self.host)

        self.enabled = True
        while self.enabled:
            # if not telnet - enable it
            if not self._check_port(23) and not self._enable_telnet():
                time.sleep(30)
                continue

            devices = self._prepare_gateway(with_devices=True)
            if devices:
                self.gw_topic = f"gw/{devices[0]['mac'][2:].upper()}/"
                self.setup_devices(devices)
                break

        self.update_time_offset()
        self.mesh_start()

        while self.enabled:
            # if not telnet - enable it
            if not self._check_port(23) and not self._enable_telnet():
                time.sleep(30)
                continue

            # if not mqtt - enable it (handle Mi Home and ZHA mode)
            if not self._prepare_gateway() or not self._mqtt_connect():
                time.sleep(60)
                continue

            self.mqtt.loop_forever()

        self.debug("Stop main thread")

    def update_time_offset(self):
        gw_time = ntp_time(self.host)
        if gw_time:
            self.time_offset = gw_time - time.time()
            self.debug(f"Gateway time offset: {self.time_offset}")

    def _check_port(self, port: int):
        """Check if gateway port open."""
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            return s.connect_ex((self.host, port)) == 0
        finally:
            s.close()

    def _enable_telnet(self):
        """Enable telnet with miio protocol."""
        if self.miio.send("enable_telnet_service") != "ok":
            self.debug(f"Can't enable telnet")
            return False
        return True

    def _prepare_gateway(self, with_devices: bool = False):
        """Launching the required utilities on the hub, if they are not already
        running.
        """
        self.debug("Prepare Gateway")
        try:
            shell = TelnetShell(self.host)

            if self.ver is None:
                self.ver = shell.get_version()
                self.debug(f"Version: {self.ver}")

            ps = shell.get_running_ps()

            if "mosquitto -d" not in ps:
                self.debug("Run public mosquitto")
                shell.run_public_mosquitto()

            if "ntpd" not in ps:
                # run NTPd for sync time
                shell.run_ntpd()

            # all data or only necessary events
            pattern = (
                '\\{"'
                if "miio" in self._debug
                else "ble_event|properties_changed|heartbeat"
            )

            if f"awk /{pattern} {{" not in ps:
                self.debug(f"Redirect miio to MQTT")
                shell.redirect_miio2mqtt(pattern, self.ver_miio)

            if self.options.get("buzzer"):
                if "dummy:basic_gw" not in ps:
                    self.debug("Disable buzzer")
                    shell.stop_buzzer()
            else:
                if "dummy:basic_gw" in ps:
                    self.debug("Enable buzzer")
                    shell.run_buzzer()

            if self.options.get("zha"):
                if "socat" not in ps:
                    if "Received" in shell.check_or_download_socat():
                        self.debug("Download socat")
                    self.debug("Run socat")
                    shell.run_socat()

                if "Lumi_Z3GatewayHost_MQTT" in ps:
                    self.debug("Stop Lumi Zigbee")
                    shell.stop_lumi_zigbee()

            else:
                if "socat" in ps:
                    self.debug("Stop socat")
                    shell.stop_socat()

                if (
                    self.parent_scan_interval >= 0
                    and "Lumi_Z3GatewayHost_MQTT -n 1 -b 115200 -l" not in ps
                ):
                    self.debug("Run public Zigbee console")
                    shell.run_public_zb_console(self.ver_z3)

                elif "Lumi_Z3GatewayHost_MQTT" not in ps:
                    self.debug("Run Lumi Zigbee")
                    shell.run_lumi_zigbee()

            if with_devices:
                self.debug("Get devices")
                return self._get_devices(shell)

            return True

        except (ConnectionRefusedError, socket.timeout):
            return False

        except Exception as e:
            self.debug(f"Can't read devices: {e}")
            return False

    def _mqtt_connect(self) -> bool:
        try:
            self.mqtt.reconnect()
            return True
        except Exception:
            return False

    def _miio_connect(self) -> bool:
        if not self.miio.ping():
            self.debug("Can't send handshake")
            return False

        return True

    def _get_devices(self, shell: TelnetShell):
        """Load devices info for Coordinator, Zigbee and Mesh."""

        # 1. Read coordinator info
        raw = shell.read_file("/data/miio/device.conf").decode()
        m = re.search(r"did=(\d+)", raw)

        raw = shell.read_file("/data/zigbee/coordinator.info")
        device = json.loads(raw)
        devices = [
            {
                "did": "lumi.0",
                "model": "lumi.gateway.mgl03",
                "mac": device["mac"],
                "type": "gateway",
                "init": {
                    "firmware lock": shell.check_firmware_lock(),
                    "alarm_did": m[1],
                },
            }
        ]

        # 2. Read zigbee devices
        if not self.options.get("zha"):
            # read Silicon devices DB
            nwks = {}
            try:
                raw = shell.read_file("/data/silicon_zigbee_host/devices.txt")
                raw = raw.decode().split(" ")
                for i in range(0, len(raw) - 1, 32):
                    ieee = reversed(raw[i + 3 : i + 11])
                    ieee = "".join(f"{i:>02s}" for i in ieee)
                    nwks[ieee] = f"{raw[i]:>04s}"
            except Exception:
                _LOGGER.exception("Can't read Silicon devices DB")

            # read Xiaomi devices DB
            raw = shell.read_file(
                "/data/zigbee_gw/" + self.ver_zigbee_db, as_base64=True
            )
            if raw.startswith(b"unqlite"):
                db = Unqlite(raw)
                data = db.read_all()
            else:
                raw = re.sub(br"}\s+{", b",", raw)
                data = json.loads(raw)

            # data = {} or data = {'dev_list': 'null'}
            dev_list = json.loads(data.get("dev_list", "null")) or []

            for did in dev_list:
                model = data[did + ".model"]
                desc = utils.get_device(model)

                # skip unknown model
                if desc is None:
                    self.debug(f"{did} has an unsupported modell: {model}")
                    continue

                retain = json.loads(data[did + ".prop"])["props"]
                self.debug(f"{did} {model} retain: {retain}")

                params = {
                    p[2]: retain.get(p[1])
                    for p in (desc["params"] or desc["mi_spec"])
                    if p[1] is not None
                }

                ieee = f"{data[did + '.mac']:>016s}"
                device = {
                    "did": did,
                    "mac": "0x" + data[did + ".mac"],
                    "ieee": ieee,
                    "nwk": nwks.get(ieee),
                    "model": data[did + ".model"],
                    "type": "zigbee",
                    "zb_ver": data[did + ".version"],
                    "init": utils.fix_xiaomi_props(params),
                    "online": retain.get("alive", 1) == 1,
                }
                devices.append(device)

        # 3. Read bluetooth devices
        if self._ble:
            raw = shell.read_file("/data/miio/mible_local.db", as_base64=True)
            db = SQLite(raw)

            # load BLE devices
            rows = db.read_table("gateway_authed_table")
            for row in rows:
                device = {
                    "did": row[4],
                    "mac": RE_REVERSE.sub(r"\6\5\4\3\2\1", row[1]),
                    "model": row[2],
                    "type": "ble",
                }
                devices.append(device)

            # load Mesh groups
            try:
                mesh_groups = {}

                rows = db.read_table(self.ver_mesh_group)
                for row in rows:
                    # don't know if 8 bytes enougth
                    mac = int(row[0]).to_bytes(8, "big").hex()
                    device = {
                        "did": "group." + row[0],
                        "mac": mac,
                        "model": 0,
                        "childs": [],
                        "type": "mesh",
                    }
                    group_addr = row[1]
                    mesh_groups[group_addr] = device

                # load Mesh bulbs
                rows = db.read_table("mesh_device")
                for row in rows:
                    device = {
                        "did": row[0],
                        "mac": row[1].replace(":", ""),
                        "model": row[2],
                        "type": "mesh",
                    }
                    devices.append(device)

                    group_addr = row[5]
                    if group_addr in mesh_groups:
                        # add bulb to group if exist
                        mesh_groups[group_addr]["childs"].append(row[0])

                for device in mesh_groups.values():
                    if device["childs"]:
                        devices.append(device)

            except Exception:
                _LOGGER.exception("Can't read mesh devices")

        # for testing purposes
        for k, v in self.default_devices.items():
            if k[0] == "_":
                devices.append(v)

        return devices

    def lock_firmware(self, enable: bool):
        self.debug(f"Set firmware lock to {enable}")
        try:
            shell = TelnetShell(self.host)
            if "Received" in shell.check_or_download_busybox():
                self.debug("Download busybox")
            shell.lock_firmware(enable)
            locked = shell.check_firmware_lock()
            shell.close()
            return enable == locked

        except Exception as e:
            self.debug(f"Can't set firmware lock: {e}")
            return False

    def on_connect(self, client, userdata, flags, rc):
        self.debug("MQTT connected")
        self.mqtt.subscribe("#")

        self.available = True
        self.process_gw_stats()

    def on_disconnect(self, client, userdata, rc):
        self.debug("MQTT disconnected")
        # force end mqtt.loop_forever()
        self.mqtt.disconnect()

        self.available = False
        self.process_gw_stats()

    def on_message(self, client: Client, userdata, msg: MQTTMessage):
        try:
            topic = msg.topic

            if "mqtt" in self._debug:
                self.debug(f"[MQ] {topic} {msg.payload.decode()}")

            if topic == "zigbee/send":
                payload = json.loads(msg.payload)
                self.process_message(payload)

            elif topic == "log/miio":
                for raw in utils.extract_jsons(msg.payload):
                    if self._ble and b"_async.ble_event" in raw:
                        data = json.loads(raw)["params"]
                        self.process_ble_event(data)
                        self.process_ble_stats(data)
                    elif self._ble and b"properties_changed" in raw:
                        data = json.loads(raw)["params"]
                        self.debug(f"Process props {data}")
                        self.process_mesh_data(data)
                    elif b"event.gw.heartbeat" in raw:
                        payload = json.loads(raw)["params"][0]
                        self.process_gw_stats(payload)
                        # time offset may changed right after gw.heartbeat
                        self.update_time_offset()

            elif topic == "log/z3":
                self.process_z3(msg.payload.decode())

            elif topic.endswith("/heartbeat"):
                payload = json.loads(msg.payload)
                self.process_gw_stats(payload)

            elif topic.endswith(("/MessageReceived", "/devicestatechange")):
                payload = json.loads(msg.payload)
                self.process_zb_stats(payload)

            # read only retained ble
            elif topic.startswith("ble") and msg.retain:
                payload = json.loads(msg.payload)
                self.process_ble_retain(topic[4:], payload)

            elif self.pair_model and topic.endswith("/commands"):
                self.process_pair(msg.payload)

        except Exception:
            _LOGGER.exception(f"Processing MQTT: {msg.topic} {msg.payload}")

    def setup_devices(self, devices: list):
        """Add devices to hass."""
        for device in devices:
            if device["type"] in ("gateway", "zigbee"):
                desc = utils.get_device(device["model"])
                if not desc:
                    self.debug(f"Unsupported model: {device}")
                    continue

                self.debug(f"Setup Zigbee device {device}")

                device.update(desc)

                # update params from config
                default_config = self.default_devices.get(
                    device["mac"]
                ) or self.default_devices.get(device["did"])
                if default_config:
                    device.update(default_config)

                self.devices[device["did"]] = device

                for param in device["params"] or device["mi_spec"]:
                    domain = param[3]
                    if not domain:
                        continue

                    # wait domain init
                    while domain not in self.setups:
                        time.sleep(1)

                    self.setups[domain](self, device, param[2])

            elif device["type"] == "mesh":
                desc = bluetooth.get_device(device["model"], "Mesh")
                device.update(desc)

                self.debug(f"Setup Mesh device {device}")

                # update params from config
                default_config = self.default_devices.get(device["did"])
                if default_config:
                    device.update(default_config)

                device["online"] = False

                self.devices[device["did"]] = device

                for param in device["params"]:
                    domain = param[3]
                    if not domain:
                        continue

                    # wait domain init
                    while domain not in self.setups:
                        time.sleep(1)

                    self.setups[domain](self, device, param[2])

            elif device["type"] == "ble":
                # only save info for future
                desc = bluetooth.get_device(device["model"], "BLE")
                device.update(desc)

                # update params from config
                default_config = self.default_devices.get(device["did"])
                if default_config:
                    device.update(default_config)

                self.devices[device["did"]] = device

                device["init"] = {}

            if self.options.get("stats") and device["type"] != "mesh":
                while "sensor" not in self.setups:
                    time.sleep(1)
                self.setups["sensor"](self, device, device["type"])

    def process_message(self, data: dict):
        if data["cmd"] == "heartbeat":
            # don't know if only one item
            assert len(data["params"]) == 1, data

            data = data["params"][0]
            pkey = "res_list"
        elif data["cmd"] == "report":
            pkey = "params" if "params" in data else "mi_spec"
        elif data["cmd"] in ("write_rsp", "read_rsp"):
            pkey = "results"
        elif data["cmd"] == "write_ack":
            return
        else:
            _LOGGER.warning(f"Unsupported cmd: {data}")
            return

        did = data["did"]

        # skip without callback
        if did not in self.updates:
            return

        ts = time.time()

        device = self.devices[did]
        payload = {}

        # convert codes to names
        for param in data[pkey]:
            if param.get("error_code", 0) != 0:
                continue

            if "res_name" in param:
                prop = param["res_name"]
            elif "piid" in param:
                prop = f"{param['siid']}.{param['piid']}"
            elif "eiid" in param:
                prop = f"{param['siid']}.{param['eiid']}"
            else:
                _LOGGER.warning(f"Unsupported param: {data}")
                return

            if prop in GLOBAL_PROP:
                prop = GLOBAL_PROP[prop]
            else:
                prop = next(
                    (
                        p[2]
                        for p in (device["params"] or device["mi_spec"])
                        if p[0] == prop
                    ),
                    prop,
                )

            # https://github.com/Koenkk/zigbee2mqtt/issues/798
            # https://www.maero.dk/aqara-temperature-humidity-pressure-sensor-teardown/
            if prop == "temperature":
                if -4000 < param["value"] < 12500:
                    payload[prop] = param["value"] / 100.0
            elif prop == "humidity":
                if 0 <= param["value"] <= 10000:
                    payload[prop] = param["value"] / 100.0
            elif prop == "pressure":
                payload[prop] = param["value"] / 100.0
            elif prop == "battery" and param["value"] > 1000:
                # xiaomi light sensor
                payload[prop] = round((min(param["value"], 3200) - 2500) / 7)
            elif prop == "alive" and param["value"]["status"] == "offline":
                device["online"] = False
            elif prop == "angle":
                # xiaomi cube 100 points = 360 degrees
                payload[prop] = param["value"] * 4
            elif prop == "duration":
                # xiaomi cube
                payload[prop] = param["value"] / 1000.0
            elif prop in ("consumption", "power"):
                payload[prop] = round(param["value"], 2)
            elif "value" in param:
                payload[prop] = param["value"]
            elif "arguments" in param:
                if prop == "motion":
                    payload[prop] = 1
                else:
                    payload[prop] = param["arguments"]

        # no time in device add command
        ts = (
            round(ts - data["time"] * 0.001 + self.time_offset, 2)
            if "time" in data
            else "?"
        )
        self.debug(f"{device['did']} {device['model']} <= {payload} [{ts}]")

        if payload:
            device["online"] = True

        for handler in self.updates[did]:
            handler(payload)

        if "added_device" in payload:
            # {'did': 'lumi.fff', 'mac': 'fff', 'model': 'lumi.sen_ill.mgl01',
            # 'version': '21', 'zb_ver': '3.0'}
            device = payload["added_device"]
            device["mac"] = "0x" + device["mac"]
            device["type"] = "zigbee"
            device["init"] = payload
            self.setup_devices([device])

    def process_ble_event(self, data: dict):
        self.debug(f"Process BLE {data}")

        pdid = data["dev"].get("pdid")

        did = data["dev"]["did"]
        if did not in self.devices:
            mac = (
                data["dev"]["mac"].replace(":", "").lower()
                if "mac" in data["dev"]
                else "ble_" + did.replace("blt.3.", "")
            )
            self.devices[did] = device = {
                "did": did,
                "mac": mac,
                "init": {},
                "type": "bluetooth",
            }
            desc = bluetooth.get_device(pdid, "BLE")
            device.update(desc)

            # update params from config
            default_config = self.default_devices.get(did)
            if default_config:
                device.update(default_config)

        else:
            device = self.devices[did]

        if isinstance(data["evt"], list):
            # check if only one
            assert len(data["evt"]) == 1, data
            payload = bluetooth.parse_xiaomi_ble(data["evt"][0], pdid)
        elif isinstance(data["evt"], dict):
            payload = bluetooth.parse_xiaomi_ble(data["evt"], pdid)
        else:
            payload = None

        if payload is None:
            self.debug(f"Unsupported BLE {data}")
            return

        # init entities if needed
        init = device["init"]
        for k in payload.keys():
            if k in init:
                # update for retain
                init[k] = payload[k]
                continue

            init[k] = payload[k]

            domain = bluetooth.get_ble_domain(k)
            if not domain:
                continue

            # wait domain init
            while domain not in self.setups:
                time.sleep(1)

            self.setups[domain](self, device, k)

        if did in self.updates:
            for handler in self.updates[did]:
                handler(payload)

        raw = json.dumps(init, separators=(",", ":"))
        self.mqtt.publish(f"ble/{did}", raw, retain=True)

    def process_ble_retain(self, did: str, payload: dict):
        if did not in self.devices:
            self.debug(f"BLE device {did} is no longer on the gateway")
            return

        self.debug(f"{did} retain: {payload}")

        device = self.devices[did]

        # init entities if needed
        for k in payload.keys():
            # don't retain action and motion
            if k in device["init"]:
                continue

            if k in ("action", "motion"):
                device["init"][k] = ""
            else:
                device["init"][k] = payload[k]

            domain = bluetooth.get_ble_domain(k)
            if not domain:
                continue

            # wait domain init
            while domain not in self.setups:
                time.sleep(1)

            self.setups[domain](self, device, k)

        if did in self.updates:
            for handler in self.updates[did]:
                handler(payload)

    def process_pair(self, raw: bytes):
        # get shortID and eui64 of paired device
        if b"lumi send-nwk-key" in raw:
            # create model response
            payload = (
                f"0x18010105000042{len(self.pair_model):02x}"
                f"{self.pair_model.encode().hex()}"
            )
            m = RE_NWK_KEY.search(raw.decode())
            self.pair_payload = json.dumps(
                {
                    "sourceAddress": m[1],
                    "eui64": "0x" + m[2],
                    "profileId": "0x0104",
                    "clusterId": "0x0000",
                    "sourceEndpoint": "0x01",
                    "destinationEndpoint": "0x01",
                    "APSCounter": "0x01",
                    "APSPlayload": payload,
                },
                separators=(",", ":"),
            )

        # send model response "from device"
        elif b"zdo active " in raw:
            self.mqtt.publish(self.gw_topic + "MessageReceived", self.pair_payload)

    def send(self, device: dict, data: dict):
        payload = {"cmd": "write", "did": device["did"]}

        # convert hass prop to lumi prop
        if device["mi_spec"]:
            params = []
            for k, v in data.items():
                if k == "switch":
                    v = bool(v)
                k = next(p[0] for p in device["mi_spec"] if p[2] == k)
                params.append({"siid": k[0], "piid": k[1], "value": v})

            payload["mi_spec"] = params
        else:
            params = [
                {
                    "res_name": next(p[0] for p in device["params"] if p[2] == k),
                    "value": v,
                }
                for k, v in data.items()
            ]

            payload = {
                "cmd": "write",
                "did": device["did"],
                "params": params,
            }

        self.debug(f"{device['did']} {device['model']} => {payload}")
        payload = json.dumps(payload, separators=(",", ":")).encode()
        self.mqtt.publish("zigbee/recv", payload)

    def send_telnet(self, *args: str):
        try:
            shell = TelnetShell(self.host)
            for command in args:
                if command == "ftp":
                    shell.check_or_download_busybox()
                    shell.run_ftp()
                else:
                    shell.exec(command)
            shell.close()

        except Exception as e:
            _LOGGER.exception(f"Telnet command error: {e}")

    def send_mqtt(self, cmd: str):
        if cmd == "publishstate":
            self.mqtt.publish(self.gw_topic + "publishstate")

    def get_device(self, mac: str) -> Optional[dict]:
        for device in self.devices.values():
            if device.get("mac") == mac:
                return device
        return None


def is_gw3(host: str, token: str) -> Optional[str]:
    device = SyncmiIO(host, token)
    info = device.info()

    if not info:
        return "cant_connect"

    if info["model"] != "lumi.gateway.mgl03":
        return "wrong_model"

    return None


def get_lan_key(device: dict):
    device = SyncmiIO(device["localip"], device["token"])
    resp = device.send("get_lumi_dpf_aes_key")
    if resp is None:
        return "Can't connect to gateway"
    if len(resp[0]) == 16:
        return resp[0]
    key = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(16)
    )
    resp = device.send("set_lumi_dpf_aes_key", [key])
    if resp[0] == "ok":
        return key
    return "Can't update gateway key"



import functools

import pytest
from fhempy.lib import utils


def test_local_ip():
    ip = utils.get_local_ip()
    assert ip != "127.0.0.1"


def test_encrypt_decrypt():
    teststring = "This is a test string"
    fhem_unique_id = "a3e36c8ec8622a0de0e11191dc430a34"
    encrypted_string = utils.encrypt_string(teststring, fhem_unique_id)
    decrypted_string = utils.decrypt_string(encrypted_string, fhem_unique_id)
    assert teststring == decrypted_string
    return decrypted_string


@pytest.mark.asyncio
async def test_run_blocking():
    test_string = await utils.run_blocking(functools.partial(test_encrypt_decrypt))
    assert "This is a test string" == test_string


@pytest.mark.asyncio
async def test_run_blocking_task():
    foo = None

    def set_foo():
        nonlocal foo
        foo = "bar"

    task = utils.run_blocking_task(functools.partial(set_foo))
    await task
    assert foo == "bar"


@pytest.mark.asyncio
async def test_handle_define_attr(mocker):
    async def setDevAttrList(hashname, attrlist):
        assert hashname == "test"
        assert str(attrlist) == "attr1 attr2 attr3 attr4 attr5:on,off,test"

    async def AttrVal(hashname, attr, default):
        if attr == "attr3":
            return "test33"
        return ""

    mocker.patch("fhempy.lib.fhem.setDevAttrList", setDevAttrList)
    mocker.patch("fhempy.lib.fhem.AttrVal", AttrVal)

    class TestClass:
        async def set_attr(self, hash):
            assert self._attr_attr4 == "test4"

    testinstance = TestClass()
    attr_conf = {
        "attr1": {},
        "attr2": {"default": 1, "format": "int"},
        "attr3": {"default": "test3"},
        "attr4": {"default": "test4", "function": "set_attr"},
        "attr5": {"default": "test5", "options": "on,off,test"},
    }
    hash = {"NAME": "test"}
    await utils.handle_define_attr(attr_conf, testinstance, hash)
    assert testinstance._attr_attr1 == ""
    assert testinstance._attr_attr2 == 1
    assert testinstance._attr_attr3 == "test33"
    assert testinstance._attr_attr4 == "test4"
    assert testinstance._attr_attr5 == "test5"


@pytest.mark.asyncio
async def test_handle_attr():
    class TestClass:
        async def set_attr(self, hash):
            assert self._attr_attr4 == "asdf"

    testinstance = TestClass()
    attr_conf = {
        "attr1": {},
        "attr2": {"default": 1, "format": "int"},
        "attr3": {"default": "test3"},
        "attr4": {"default": "test4", "function": "set_attr"},
        "attr5": {"default": "test5", "options": "on,off,test"},
    }
    hash = {"NAME": "test"}
    # set new value
    await utils.handle_attr(
        attr_conf, testinstance, hash, ["set", "test", "attr1", "NEWVALUE"], {}
    )
    assert testinstance._attr_attr1 == "NEWVALUE"

    # set integer
    await utils.handle_attr(
        attr_conf, testinstance, hash, ["set", "test", "attr2", "2"], {}
    )
    assert testinstance._attr_attr2 == 2

    # del attribute
    await utils.handle_attr(
        attr_conf, testinstance, hash, ["del", "test", "attr2", ""], {}
    )
    assert testinstance._attr_attr2 == 1

    await utils.handle_attr(
        attr_conf, testinstance, hash, ["set", "test", "attr4", "asdf"], {}
    )
    assert testinstance._attr_attr4 == "asdf"


def test_flatten_json():
    json = {"asdf": {"nested": {"nested2": "x"}}}
    flat_json = utils.flatten_json(json)
    for element in flat_json:
        assert element == "asdf_nested_nested2"
        assert flat_json[element] == "x"


def test_convert2format():
    listdef = {"format": "int"}
    newval = utils.convert2format("3", listdef)
    assert newval == 3

    listdef = {"format": "int"}
    newval = utils.convert2format("3", listdef)
    assert newval == 3
    assert isinstance(newval, int) == True

    listdef = {"format": "str"}
    newval = utils.convert2format("3", listdef)
    assert newval == "3"
    assert isinstance(newval, str) == True

    listdef = {"format": "float"}
    newval = utils.convert2format("3", listdef)
    assert newval == 3.0
    assert isinstance(newval, float) == True

    listdef = {}
    newval = utils.convert2format("3", listdef)
    assert newval == "3"

    listdef = {}
    newval = utils.convert2format(3, listdef)
    assert newval == 3


@pytest.mark.asyncio
async def test_handle_set():
    set_list_conf = {
        "mode": {
            "args": ["mode"],
            "argsh": ["mode"],
            "params": {"mode": {"optional": False}},
            "options": "eco,comfort",
        },
        "desiredTemp": {"args": ["temperature"], "options": "slider,10,1,30"},
        "holidayMode": {
            "args": ["temperature", "till", "end"],
            "params": {"till": {"default": "31.12.2030"}, "end": {"default": "23:59"}},
        },
        "on": {
            "args": ["seconds"],
            "params": {"seconds": {"optional": True, "format": "int"}},
        },
        "off": {},
    }

    newstate = None

    class TestClass:
        def __init__(self):
            pass

        async def set_on(self, hash, params):
            nonlocal newstate
            if len(params) == 0:
                newstate = "on"
            else:
                newstate = params

        async def set_off(self, hash, params):
            assert hash["NAME"] == "testhash"
            nonlocal newstate
            newstate = "off"

        async def set_mode(self, hash, params):
            nonlocal newstate
            newstate = params

        async def set_holidayMode(self, hash, params):
            nonlocal newstate
            newstate = params

    testhash = {"NAME": "testhash"}

    testinstance = TestClass()
    retval = await utils.handle_set(
        set_list_conf, testinstance, testhash, ["testhash"], {}
    )
    assert (
        retval
        == "Unknown argument ?, choose one of mode:eco,comfort desiredTemp:slider,10,1,30 holidayMode on off:noArg"
    )

    retval = await utils.handle_set(
        set_list_conf, testinstance, testhash, ["testhash", "?"], {}
    )
    assert (
        retval
        == "Unknown argument ?, choose one of mode:eco,comfort desiredTemp:slider,10,1,30 holidayMode on off:noArg"
    )

    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "mode"],
        {},
    )
    assert retval == "Required argument mode missing."

    retval = await utils.handle_set(
        set_list_conf, testinstance, testhash, ["testhash", "nopossiblecommand"], {}
    )
    assert retval == "Command not available for this device."

    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "off", "toomanyarguments"],
        {},
    )
    assert retval == "Too many parameters provided: toomanyarguments"

    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "off"],
        {},
    )
    assert newstate == "off"

    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "mode", "eco"],
        {},
    )
    assert newstate == {"mode": "eco"}

    newstate = None
    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "mode"],
        {"mode": "eco"},
    )
    assert newstate == {"mode": "eco"}

    newstate = None
    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "mode", "nonexistent"],
        {},
    )
    assert retval == None
    assert newstate == {"mode": "nonexistent"}

    newstate = None
    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "holidayMode", "21"],
        {},
    )
    assert retval == None
    assert newstate == {"till": "31.12.2030", "end": "23:59", "temperature": "21"}

    newstate = None
    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "holidayMode", "21", "31.12.2040", "23:38"],
        {},
    )
    assert retval == None
    assert newstate == {"till": "31.12.2040", "end": "23:38", "temperature": "21"}

    newstate = None
    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "on"],
        {},
    )
    assert retval == None
    assert newstate == "on"

    newstate = None
    retval = await utils.handle_set(
        set_list_conf,
        testinstance,
        testhash,
        ["testhash", "on", "300"],
        {},
    )
    assert retval == None
    assert newstate == {"seconds": 300}



import asyncio
import functools
import struct
import time

from .. import fhem, utils
from ..core.ble import BTLEConnection
from ..generic import FhemModule

DEFAULT_TIMEOUT = 1

HANDLE_RW_PASSWORD = 0x0048
HANDLE_R_WATERING = 0x0015
HANDLE_W_WATERING = 0x0013
HANDLE_R_BATTERY = 0x0039
HANDLE_R_TEMPERATURE = 0x003B
HANDLE_R_MIN_TEMP = 0x003D
HANDLE_R_MAX_TEMP = 0x003F
HANDLE_R_FIRMWARE = 0x004E
HANDLE_RW_DEVNAME = 0x0052
HANDLE_RW_ECO_PART1 = 0x0033
HANDLE_RW_ECO_PART2 = 0x0045
HANDLE_RW_TIME_OFFSET = 0x0035
HANDLE_R_MAC = 0x004A
HANDLE_RW_INCREASEREDUCE = 0x0043
HANDLE_W_COMMITCODE = 0x0037
HANDLE_RW_TIMERS = [
    0x0017,
    0x0019,
    0x001B,
    0x001D,
    0x001F,
    0x0021,
    0x0023,
    0x0031,
    0x0025,
    0x0027,
    0x0029,
    0x002B,
    0x002D,
    0x002F,
]


class gfprobt(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        set_conf = {
            "update": {"help": "Retrieve values from GF Pro BT"},
            "off": {"help": "Turn off watering"},
            "toggle": {"help": "Toggle on/off"},
            "on": {
                "args": ["onseconds"],
                "params": {
                    "onseconds": {"default": 0, "format": "int", "optional": True}
                },
                "help": (
                    "Turn on watering. You can use seconds as parameter to define the duration.<br>"
                    "Duration will be set directly on the device to have a safe switch off even if connectivity is broken"
                ),
            },
            "devicename": {"help": "Set the devicename of the irrigation control"},
            "adjust": {
                "help": "Adjust the timings with parameters:<br>percent<br>duration in minutes"
            },
            "eco": {"help": "Activate/decactivate eco mode"},
        }
        self.set_set_config(set_conf)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define irrigation_control PythonModule gfprobt <MAC>"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        self._conn = BTLEConnection(
            self._mac,
            keep_connected=True,
            connection_established_callback=self.write_password,
        )
        self.create_async_task(self.update())

    # write_password is running in another thread
    def write_password(self, mac):
        self._conn.write_characteristic(HANDLE_RW_PASSWORD, b"123456")

    async def set_update(self, hash, params):
        self.create_async_task(self.update())

    async def set_on(self, hash, params):
        onseconds = params["onseconds"]
        utils.run_blocking_task(functools.partial(self.blocking_on, onseconds))

    async def set_off(self, hash, params):
        utils.run_blocking_task(functools.partial(self.blocking_off))

    async def set_toggle(self, hash, params):
        utils.run_blocking_task(functools.partial(self.blocking_toggle))

    def blocking_on(self, onseconds):
        self.blocking_update_watering()
        if self._watering == 0:
            self.blocking_toggle()

    def blocking_off(self):
        self.blocking_update_watering()
        if self._watering == 1:
            self.blocking_toggle()

    def blocking_toggle(self):
        self._conn.write_characteristic(HANDLE_W_WATERING, b"\x00")
        self._conn.write_characteristic(HANDLE_W_WATERING, b"\x01")

    async def update(self):
        await utils.run_blocking(functools.partial(self.blocking_update))
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", self._watering)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "watering", self._watering)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "batteryVoltage", self._battery
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "temperature", self._temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "min_temperature", self._min_temp
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "max_temperature", self._max_temp
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "firmware", self._firmware)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "devname", self._devname)
        await fhem.readingsEndUpdate(self.hash, 1)

    def blocking_update_watering(self):
        self._watering = struct.unpack(
            "<b", self._conn.read_characteristic(HANDLE_R_WATERING)
        )[0]

    def blocking_update(self):
        self.blocking_update_watering()
        self._battery = struct.unpack(
            "<h", self._conn.read_characteristic(HANDLE_R_BATTERY)
        )[0]
        self._temperature = struct.unpack(
            "<h", self._conn.read_characteristic(HANDLE_R_TEMPERATURE)
        )[0]
        self._min_temp = struct.unpack(
            "<h", self._conn.read_characteristic(HANDLE_R_MIN_TEMP)
        )[0]
        self._max_temp = struct.unpack(
            "<h", self._conn.read_characteristic(HANDLE_R_MAX_TEMP)
        )[0]
        self._firmware = self._conn.read_characteristic(HANDLE_R_FIRMWARE)
        self._firmware = str(self._firmware[1]) + "." + str(self._firmware[0])
        self._devname = self._conn.read_characteristic(HANDLE_RW_DEVNAME).decode(
            "utf-8"
        )
        self._eco = str(self._conn.read_characteristic(HANDLE_RW_ECO_PART1))
        self._eco += " " + str(self._conn.read_characteristic(HANDLE_RW_ECO_PART2))
        self._timeoffset = struct.unpack(
            "<I", self._conn.read_characteristic(HANDLE_RW_TIME_OFFSET)
        )[0]
        self._devmac = str(self._conn.read_characteristic(HANDLE_R_MAC))
        self._increasereduce = self._conn.read_characteristic(HANDLE_RW_INCREASEREDUCE)
        self._raw_timers = {}
        for handle_timer in HANDLE_RW_TIMERS:
            self._raw_timers[handle_timer] = self._conn.read_characteristic(
                handle_timer
            )
            self.logger.debug(self._raw_timers[handle_timer])

    def write_offset(self):
        now = time.localtime()
        sec_since_mon = (
            now.tm_wday * 3600 * 24 + now.tm_hour * 3600 + now.tm_min * 60 + now.tm_sec
        )
        sec_since_mon = struct.pack("<I", sec_since_mon)
        self._conn.writeCharacteristic(HANDLE_RW_TIME_OFFSET, sec_since_mon)

    def commit_code(self):
        self._conn.writeCharacteristic(HANDLE_W_COMMITCODE, b"00")
        self._conn.writeCharacteristic(HANDLE_W_COMMITCODE, b"01")



#!/usr/bin/env python3
"""Constants for the Tuya integration."""

DOMAIN = "tuya_v2"

CONF_PROJECT_TYPE = "tuya_project_type"
CONF_ENDPOINT = "endpoint"
CONF_ACCESS_ID = "access_id"
CONF_ACCESS_SECRET = "access_secret"
CONF_USERNAME = "username"
CONF_PASSWORD = "password"
CONF_COUNTRY_CODE = "country_code"
CONF_APP_TYPE = "tuya_app_type"

TUYA_DISCOVERY_NEW = "tuya_v2_discovery_new_{}"
TUYA_DEVICE_MANAGER = "tuya_device_manager"
TUYA_MQTT_LISTENER = "tuya_mqtt_listener"
TUYA_HA_TUYA_MAP = "tuya_ha_tuya_map"
TUYA_HA_DEVICES = "tuya_ha_devices"
TUYA_SETUP_PLATFORM = "setup_platform"

TUYA_ENDPOINT = {
    "https://openapi.tuyaus.com": "America",
    "https://openapi.tuyacn.com": "China",
    "https://openapi.tuyaeu.com": "Europe",
    "https://openapi.tuyain.com": "India",
    "https://openapi-ueaz.tuyaus.com": "EasternAmerica",
    "https://openapi-weaz.tuyaeu.com": "WesternEurope",
}

TUYA_PROJECT_TYPE = {1: "Custom Development", 0: "Smart Home PaaS"}

TUYA_APP_TYPE = {"tuyaSmart": "TuyaSmart", "smartlife": "Smart Life"}

TUYA_SUPPORT_HA_TYPE = [
    "switch",
    "fan",
    "cover",
    "climate",
    "light",
    "sensor",
    "binary_sensor",
    "humidifier",
    "number"
    # 'alarm_control_panel'
]
